# $Id: SemaphoreSet.pm,v 1.1 2000/01/09 21:39:57 dave Exp $
# Copyright (C) 2000 Paul Martin and Dave Woodcock. All rights reserved.

package SemaphoreSet;

# {{{ pragmas

use strict;
use vars qw
  (
   $VERSION
  );

$VERSION = '0.01';

# }}}
# {{{ package variables

# These values are taken from /usr/include/sys/ipc_buf.h
# I should be using h2ph instead

my $IPC_PRIVATE = 00000;

# Flags for semget
my $IPC_CREAT   = 01000;
my $IPC_EXCL    = 02000;

# Ops for semctl
my $IPC_RMID    = 00000;
my $IPC_SET     = 00001;
my $IPC_STAT    = 00002;

# Flags for semop
my $IPC_NOWAIT  = 04000;
my $SEM_UNDO    = 10000;

# }}}
# {{{ new

sub new($%) {
  my $self = bless
    {'pid'      => $$, # PID of process which created this
     'nsems'    => 0,  # Number of sems in the set
     'init_val' => [], # Array of initial values
     'sem_num'  => {}, # Look up sem_num from id
    }, shift;

  # Read in the ids and initial values
  while (@_) {
    my ($id, $init_val) = (shift, shift);
    $init_val >= 0
      or die("SemaphoreSet::new: bogus initial value $init_val specified for id $id\n");
    $self->{'sem_num'}->{$id}  = $self->{'nsems'}++;
    push @{$self->{'init_val'}}, $init_val;
  }

  # Create the new semaphore set
  $self->{'semid'} = semget
    ($IPC_PRIVATE, $self->{'nsems'}, $IPC_CREAT | $IPC_EXCL | 0700);
  defined($self->{'semid'}) or die("SemaphoreSet::new: semget failed: $!\n");

  # Set the initial values
  my $semop = pack('s*', map {
    my $i = $self->{'init_val'}->[$_];
    $i ? ($_, $i, 0) : ();
  } 0 .. $self->{'nsems'} - 1);

  $semop eq ''
    or semop($self->{'semid'}, $semop)
    or die("SemaphoreSet::new: failed to set initial values: $!\n");

  $self;
}

# }}}
# {{{ DESTROY

# Remove the semaphore when the creating process exits

sub DESTROY($) {
  my $self = shift;
  $self->{'pid'} != $$
    or semctl($self->{'semid'}, 0, $IPC_RMID, 0)
    or die("SemaphoreSet::DESTROY: failed to remove semaphore: $!\n");
}

# }}}
# {{{ ops

sub op($$$) {
  my ($self, $id, $val) = @_;

  defined(my $sem_num = $self->{'sem_num'}->{$id})
    or die("SemaphoreSet::op: bad id: $id\n");

  semop($self->{'semid'}, pack('s*', $sem_num, $val, 0))
    or die("SemaphoreSet:op: semop failed: $!\n");
}

sub up($$)   {$_[0]->op($_[1], 1)}
sub down($$) {$_[0]->op($_[1], -1)}

# }}}

1;

__END__

# {{{ doco

=head1 NAME

SemaphoreSet - Perl module implementing simple (and restricted)
semaphores on top of SysV semaphores.

=head1 SYNOPSIS

  # The producer/consumer problem solved using SemaphoreSet

  my $sem = new SemaphoreSet
    ('full'  => 0,
     'empty' => 16,
     'mutex' => 1,
    );

  defined(my $pid = fork)
    or die("Fork failed: $!\n");

  if ($pid) { # Parent - produce
    while (1) {
      $sem->down('full');
      $sem->down('mutex');
      # Producer code goes here
      $sem->up('mutex');
      $sem->up('empty');
    }

  } else {    # Child - consume
    while (1) {
      $sem->down('empty');
      $sem->down('mutex');
      # Consumer code goes here
      $sem->up('mutex');
      $sem->up('full');
    }
  }

=head1 DESCRIPTION

SemaphoreSet implements a very restricted set of object-oriented
semaphore functionality.

=head1 AUTHOR

Dave Woodcock d_woodcock@yahoo.com

=head1 SEE ALSO

perl(1).

=cut

# }}}
# {{{ Local Variables

# Local Variables:
# mode: cperl
# folded-file: t
# End:

# }}}
