# $Id: Camera.pm,v 1.2 2000/11/21 00:44:52 dave Exp $
# Copyright (C) 2000 Paul Martin and Dave Woodcock. All rights reserved.

package Camera;

# A geomview object consists of:
# - a name
# - a coordinate transform

# Convention for coordinate transform is that the identity matrix
# corresponds to sitting at the origin with the x- and y-coordinates
# as view frame coordinates, looking in the direction of the
# negative z-axis

# {{{ pragmata

use strict;
use Vector;
use Matrix;

# }}}
# {{{ package lexicals

my $NAME  =      0; # object name
my $TRANSFORM =  1; # current coordinate transform
my $XVEC  =  	 2; # vector perp to axis
my $YVEC  = 	 3; # vector perp to axis
my $ZVEC  = 	 4; # axis
my $ASTEP = 	 5; # axial rate of movement
my $AINIT = 	 6; # initial axial position
my $PSTEP = 	 7; # perpendicular rate of movement
my $PINIT = 	 8; # initial perpendicular position
my $RSTEP = 	 9; # rate of rotation
my $TSTEP = 	10; # rate of motion of target (along ZVEC)
my $TICK  = 	11; # current tick
my $DIR   =     12; # snapshot dir
my $STEM  =     13; # snapshot stem
my $AMBIENT =   14; # ambient light level
my $HEADLIGHT = 15; # headlight light level

my $default_view = new Vector(0, 0, -1, 0);
# Arbitrarily choose the x-y plane
my $rot180 = Matrix::rows2mx([-1,0,0,0], [0,-1,0,0], [0,0,1,0], [0,0,0,1]);

# }}}
# {{{ pos

sub pos($) {
    my $self = shift;
    my $transform = $self->[$TRANSFORM];
    my $res = [];
    for (my $i = 0; $i < 3; ++$i) {
	$res->[$i] = $transform->[3]->[$i];
    }
    return bless $res, 'Vector';
}

# }}}
# {{{ set_pos

sub set_pos($$) {
    my ($self, $pos) = @_;
    my $transform = $self->[$TRANSFORM];
    for (my $i = 0; $i < 3; ++$i) {
	$transform->[3]->[$i] = $pos->[$i];
    }
}

# }}}
# {{{ line_of_sight

sub line_of_sight($) {
    my $self = shift;
    my $v = $default_view->mult($self->[$TRANSFORM]);
    $#$v = 2; # Truncate to a 3-vector
    return $v;
}

# }}}
# {{{ frame

# Given an axis this function returns an orthonormal frame
# with the unit vector in this direction as third component.

sub frame($$) {
    my ($self, $axis) = @_;

    $axis = $axis->unit;
    my $v = new Vector(1, 0, 0);

    if (abs($axis->dot($v)) > 0.9) {
	# Bad guess
	$v = new Vector(0, 1, 0);
	if (abs($axis->dot($v)) > 0.9) {
	    # Another bad guess
	    $v = new Vector(0, 0, 1);
	}
    }

    # Normalise
    $v = $v->minus($axis->mult($axis->dot($v)))->unit;
    # Final vector is cross product
    return ($v, $axis->cross($v), $axis);
}

# }}}
# {{{ new

# Arguments are:
# name: Geomview's name for the object (typically a camera)
# axis: the axis along which the object we are tracking is moving
#       (this vector will be normalised if not already a unit vector)
# ainit: initial axial distance
# tstep: the amount the tracked object moves in one tick
# rstep: the amount we move parallel to the axis in one tick
# pinit: initial radial distance 
# pstep: the amount we move perpendicularly in one tick
# rstep: the angle through which we rotate about the axis in one tick

sub new($%) {
    my $pkg = shift;
    my %args = @_;
    my $self = bless [], $pkg;

    my ($xvec, $yvec, $zvec) = $self->frame($args{'axis'} || new Vector(0,0,1));

    $self->[$NAME]  = $args{'name'} || 'unknown';
    $self->[$XVEC]  = $xvec;
    $self->[$YVEC]  = $yvec;
    $self->[$ZVEC]  = $zvec;
    $self->[$ASTEP] = defined($args{'astep'}) ? $args{'astep'} : 0.03;
    $self->[$AINIT] = defined($args{'ainit'}) ? $args{'ainit'} : 0.0;
    $self->[$TSTEP] = defined($args{'tstep'}) ? $args{'tstep'} : 0.0;
    $self->[$PSTEP] = defined($args{'pstep'}) ? $args{'pstep'} : 0.03;
    $self->[$RSTEP] = defined($args{'rstep'}) ? $args{'rstep'} : 0.1;
    $self->[$PINIT] = defined($args{'pinit'}) ? $args{'pinit'} : 0.1;
    $self->[$TRANSFORM] = Matrix::id(4);
    $self->set_pos(new Vector(0, 0, 1));
    $self->[$TICK] = 1;
    $self->[$DIR] = $args{'dir'} || '/tmp';
    $self->[$STEM] = $args{'stem'} || $args{'name'};
    $self->[$AMBIENT] = $args{'ambient'} || 10;
    $self->[$HEADLIGHT] = $args{'headlight'} || 10;
    return $self;
}

# }}}
# {{{ init

sub init($) {
    my $self = shift;
    my ($ambient, $headlight) = ($self->[$AMBIENT], $self->[$HEADLIGHT]);
    $ambient /= 100;
    $headlight /= 100;

    print <<EOF;
(merge-baseap appearance {
  lighting {
    ambient $ambient $ambient $ambient
    replacelights
    light {
	ambient 0.9 0.0 0.9
	color 0.75 0.75 0.75
	position 0.0 0.0 10.0 0.0
    }
    light {
	ambient 0.9 0.0 0.9
	color 0.6 0.6 0.6
	position 0.0 1.0 -1.0 0.0
    }
    light {
	ambient 0.9 0.9 0.9
	color 0.4 0.4 0.4
	position 1.0 -2.0 -1.0 0.0
    }
    light {                       
        color $headlight $headlight $headlight
        position  0.0 0.0 0.5 1
        location camera
    }
}})
EOF
}

# }}}
# {{{ normalise

# Over time the transform can accumulate errors which render it
# non-orthonormal. Correct for this.

sub normalise($) {
    my $mx = $_[0]->[$TRANSFORM];
    my $v1 = bless $mx->[0], 'Vector';
    my $v2 = bless $mx->[1], 'Vector';
    my $v3 = bless $mx->[2], 'Vector';

    $v1 = $v1->unit;
    $v2 = $v2->minus($v1->mult($v1->dot($v2)))->unit;
    $v3 = $v3->minus($v1->mult($v1->dot($v3)))->minus($v2->mult($v2->dot($v3)))->unit;

    $mx->[0] = $v1;
    $mx->[1] = $v2;
    $mx->[2] = $v3;
}

# }}}
# {{{ move_view

# Change transform so that the object is at $pos looking towards
# $view, and so that the change in frame between before and
# after is minimised.

sub move_view($$$) {
    my ($self, $pos, $view) = @_;
    my $old_sight = $self->line_of_sight;   # unit vector along old line of sight
    my $new_sight = $view->minus($pos)->unit; # unit vector along new line of sight
    my $perp = $old_sight->cross($new_sight);
    my $cos = $old_sight->dot($new_sight);
    my $c2 = 1 - $cos * $cos;
    my $sin = $c2 < 0.0 ? 0.0 : sqrt($c2);
    
    # If the sine is very small we're near enough parallel
    if (abs($sin) < 0.0001) {
	if ($cos < 0) {
	    # Pointing in opposite directions - need to turn around,
	    # so do a 180 degree rotation
	    $self->[$TRANSFORM] = $self->[$TRANSFORM]->mult($rot180);
	}
	$self->set_pos($pos);
	return;
    }

    # Build a new orthonormal frame in which the transform is
    # a simple rotation.
    $perp = $perp->unit;
    $new_sight = $new_sight->minus($old_sight->mult($cos))->unit;

    # $p is coordinate transform to the new frame
    my $p = Matrix::rows2mx($old_sight, $new_sight, $perp);
    # $q is the transform in the coordinates of the new frame
    my $q = Matrix::rows2mx([$cos, $sin, 0], [-$sin, $cos, 0], [0,0,1]);
    # The matrix we want is $p^T * $q * $p (since $p is orthogonal)
    $self->[$TRANSFORM]
	= $self->[$TRANSFORM]->mult(Matrix::conv3to4($p->transpose->mult($q->mult($p)), $pos));
    $self->set_pos($pos);
    $self->normalise;
}

# }}}
# {{{ tick

# Move the object forward to the next position

sub tick($) {
    my ($self) = @_;
    my $tick = $self->[$TICK]++;
    my $view = $self->[$ZVEC]->mult($self->[$TSTEP] * $tick);
    my $r = $self->[$PINIT] + $tick * $self->[$PSTEP];
    my $theta = $tick * $self->[$RSTEP];
    my $pos = $self->[$ZVEC]->mult($self->[$AINIT] + $tick * $self->[$ASTEP]);
    $pos = $pos->plus($self->[$XVEC]->mult($r * cos($theta)));
    $pos = $pos->plus($self->[$YVEC]->mult($r * sin($theta)));
    $self->move_view($pos, $view);
}

# }}}
# {{{ xform_set

# Return GCL to set the transform for the object

sub xform_set($) {
    my $self = shift;
    join('',
	 '(xform-set ',
	 $self->[$NAME],
	 " transform \n",
	 $self->[$TRANSFORM]->render,
	 ')',
	 );
}

# }}}
# {{{ gifsnap

# Write camera's current view to a gif file

sub gifsnap($) {
    my $self = shift;
    join('',
	 '(snapshot ', $self->[$NAME],
	 ' |ppmtogif>',
	 $self->[$DIR], '/', $self->[$STEM],
	 '.', sprintf("%04d", $self->[$TICK]),
	 '.gif)',
	 );
}

# }}}
# {{{ print

# For debug only

sub print($) {
    my $self = shift;
    $self->[$TRANSFORM]->print;
}

# }}}

1;

__END__

# {{{ Local Variables

# Local Variables:
# mode: perl
# folded-file: t
# End:

# }}}
