# Copyright (C) 2000 Paul Martin and Dave Woodcock. All rights reserved.
# $Id: Matrix.pm,v 1.1 2000/11/13 23:59:47 dave Exp $

use strict;
package Matrix;

# Store matrices as vectors of ROWS. Thus for a matrix $a, its height
# is @$a and its width @{$a->[0]}.

# {{{ new

# Zero Matrix with given height and width
sub new($$$) {
    my ($pkg, $height, $width) = @_;
    return bless [map {[map {0.0} 0 .. $width - 1]} 0 .. $height - 1], $pkg;
}

# }}}
# {{{ id

sub id($) {
    my $n = shift;
    return bless
	[
	 map {
	     my $i = $_;
	     [map {$i == $_ ? 1.0 : 0.0} 0 .. $n - 1]
	 } 0 .. $n - 1
	], 'Matrix';
}

# }}}
# {{{ rows2mx

# Turn a number of rows into a matrix

sub rows2mx(@) {bless [@_], 'Matrix'}

# }}}
# {{{ cols2mx

# Turn a number of columns into a matrix

sub cols2mx(@) {
    my @a = @_;
    my ($width, $height) =  (scalar(@a), scalar(@{$a[0]}));
    bless
	[
	 map {
	     my $row = $_;
	     [ map {$a[$_]->[$row]} 0 .. $width - 1 ]
	 } 0 .. $height - 1
	], 'Matrix';
}

# }}}
# {{{ mult

sub mult($$) {
    my ($a, $b) =  @_;
    my $m = [];

    for (my $i = 0; $i < @$a; ++$i) {
	for (my $k = 0; $k < @{$b->[0]}; ++$k) {
	    my $s = 0.0;
	    for (my $j = 0; $j < @$b; ++$j) {
		$s += $a->[$i]->[$j] * $b->[$j]->[$k];
	    }
	    $m->[$i]->[$k] = $s;
	}
    }
    return bless $m, 'Matrix';
}

# }}}
# {{{ transpose

sub transpose($) {
    my ($a) =  @_;
    my $m = [];

    for (my $i = 0; $i < @$a; ++$i) {
	for (my $j = 0; $j < @{$a->[0]}; ++$j) {
	    $m->[$j]->[$i] = $a->[$i]->[$j];
	}
    }
    return bless $m, 'Matrix';
}

# }}}
# {{{ render

# Render as geomview matrix

sub render($) {
    my ($a) = @_;
    my $s = '';
    for (my $i = 0; $i < @$a; ++$i) {
	$s .= sprintf("%8.4f %8.4f %8.4f %8.4f\n", @{$a->[$i]});
    }
    return $s;
}

# }}}
# {{{ conv3to4

# convert a 3x3 matrix m and 3 vector v
# to a 4x4 matrix, in the canonical fashion
# (for multiplication on the right)

sub conv3to4($$) {
    my ($m, $v) = @_;
    my $res;
    for (my $i = 0; $i < 3; ++$i) {
	for (my $j = 0; $j < 3; ++$j) {
	    $res->[$i]->[$j] = $m->[$i]->[$j];
	}
    }

    for (my $i = 0; $i < 3; ++$i) {
	$res->[3]->[$i] = $v->[$i];
	$res->[$i]->[3] = 0;
    }

    $res->[3]->[3] = 1.0;
    return bless $res, 'Matrix';
}

# }}}
# {{{ print

# Intended for debug only

sub print($) {
    my ($m) = @_;
    for (my $i = 0; $i < @$m; ++$i) {
	for (my $j = 0; $j < @{$m->[0]}; ++$j) {
	    printf("%8.4f ", $m->[$i]->[$j]);
	}
	print "\n";
    }
}

# }}}

1;

__END__

# {{{ Local Variables

# Local Variables:
# mode: perl
# folded-file: t
# End:

# }}}
