# $Id: Tilter.pm,v 1.3 2000/11/21 00:44:53 dave Exp $
# Copyright (C) 2000 Paul Martin and Dave Woodcock. All rights reserved.

# This package forks. One process bootstraps the C++ tilting
# code and churns out tilting modules. The other process controls
# the geomview camera.
#
# The two process form an instance of the classic producer/consumer
# problem, and communicate using semaphores.
#
# We communicate with geomview by writing gcl to stdout, so watch
# what you print! (If you need debug, put it on stderr.)

package Tilter;

# {{{ pragmas

use strict;
use Getopt::Std;
use SemaphoreSet;

# }}}
# {{{ signal handling

my $handler = sub {
  die "Caught SIG$_[0]\n";
};

%SIG = (
   'HUP'  => $handler, # send hang-up to all processes of terminal
   'INT'  => $handler, # interrupt key
   'QUIT' => $handler, # quit key
   'ILL'  => $handler, # illegal instruction
   'TRAP' => $handler, # debugger
   'ABRT' => $handler, # abort
   'IOT'  => $handler, # I/O trap
   'BUS'  => $handler, # bus error
   'FPE'  => $handler, # floating-point exception
   'SEGV' => $handler, # seg violation
   'TERM' => $handler, # software termination
   'XCPU' => $handler, # CPU limit exceeded
   'XFSZ' => $handler, # file size limit exceeded
   'PWR'  => $handler, # power failure restart
  );

# }}}
# {{{ package variables

my $frames;    	                         # how many frames to run for
my $buflen   	 = 32;                   # number of frames to buffer
my $dir      	 = "/tmp/tilt.$<";       # temporary directory
my $filename 	 = "$dir/cmds.$$";
my $producer_tmp = "$dir/producer_tmp.$$";
my $consumer_tmp = "$dir/consumer_tmp.$$";
my $parent_pid   = $$;
my $tidy_up      = 1;

my $sem          = new SemaphoreSet
  ('full'        => 0,
   'empty'       => $buflen,
  );

# }}}
# {{{ next_state

# Retrieve the next state from the producer

my $next_item;
my $frames_read;

sub next_state() {
    $sem->down('full');
    rename("$filename.$next_item", $consumer_tmp)
	or die("consumer failed to rename $filename.$next_item to $consumer_tmp: $!\n");
    $sem->up('empty');

    open(IN, "<$consumer_tmp")  or die("Failed to open $consumer_tmp: $!\n");
    my $state = <IN>;
    close(IN);
    ++$next_item >= $buflen and $next_item = 0;
    ++$frames_read;
    return $state;
}

# }}}
# {{{ consumer

# The consumer reads tilting characters from the producer and sends
# them to geomview. It also moves the camera between ticks.
# If the producer lags behind (which happens when we head further
# out into the dominant region), the times between displays may
# become uneven.

sub consumer(%) {
    my %args = @_;
    require 'MTime.pm';  # defines timed_exec
    require 'Camera.pm'; # module wrapping camera ops
    require 'Vector.pm'; # defines perl vector ops

    my $pause    = $args{'pause'} || 100; 
    my $step     = $args{'step'} || 40;
    my $count    = 0;
    $next_item   = 0;
    $frames_read = 0;
    my $axis = $args{'axis'} || [0.44, 0.35, 0.85];
    $axis = new Vector(@$axis);

    # Define the camera track
    my $c0 = new Camera
	('name'  => $args{'name'} || 'c0',  # Name of the camera
	 'axis'  => $axis,                  # Tracking Axis
	 'astep' => $args{'astep'} / $step, # Axial step
	 'ainit' => $args{'ainit'},         # Axial Init
	 'tstep' => $args{'tstep'} / $step, # Track step
	 'pstep' => $args{'pstep'} / $step, # Perpendicular step
	 'pinit' => $args{'pinit'},         # Perpendicular init
	 'rstep' => $args{'rstep'} / $step, # Rotation step
	 'dir'   => $args{'dir'},           # Snapshot directory
	 'stem'  => $args{'stem'},          # Snapshot stem
	 'ambient' => $args{'ambient'},   # ambient light level
	 'headlight' => $args{'headlight'}, # headlight level
	 );
    $c0->init;

    # If filming is specified and output dir non-existant,
    # create it.
    if ($args{'film'} and ! -d $args{'dir'}) {
	system('mkdir', '-p', $args{'dir'}) == 0
	    or die("Failed to mkdir $args{'dir'}: $!\n");
    }

    $| = 1;           # autoflush on 
    local $/ = undef; # slurp whole files in a single gulp

    my ($next_state, $cam_transf);

    my $doit = sub {
	print $c0->gifsnap, "\n" if $args{'film'};
	$c0->tick;
	$cam_transf = $c0->xform_set;
	$next_state = $count % $step == 0 ? next_state : '';
	print $cam_transf, "\n", $next_state, "\n";
    };

    while($frames_read < $frames) {
	if ($count == 0) {
	    # First time through just read from the producer
	    print next_state(), "\n";
	} else {
	    # Otherwise, we only go to the producer every $step ticks.
	    MTime::timed_exec($pause, $doit);
	}
	++$count;
    }
}

# }}}
# {{{ producer

# The producer does all the work. It computes characters as fast as it
# can, buffering the output in a sequence of files to be read by the
# consumer.

sub producer(%) {
    my %args = @_;

    require 'TiltingWalk.pm'; # bootstraps the C++ tilting module code
    my $root_system = $args{'root_system'} || 'B3';
    my $transient   = $args{'transient'} || [];
    my $periodic    = $args{'periodic'} || [0, 2, 1, 3, 2, 1, 3];
    my $walk        = new directedTiltingWalk($root_system, $transient, $periodic);

    my ($count, $next_item, $geom) = (0, 0);

    while ($count++ < $frames) {
	# The fact that the writes to $producer_tmp below happen in a
	# block of their own ensures that $tmp_conn is closed
	# before we rename $producer_tmp below.
	my $tmp_conn = new ostream($producer_tmp);

	if ($count == 1) {
	    # On the first pass we print the axes, initialise the 
	    # tilting module geometry, and transform the coordinates
	    # of each of these to Euclidean form. (The tilting module
	    # code works internally using the natural basis of the 
	    # root system.)
	    if ($args{'axes'}) {
		my $axes = new gvGeom('axes', $tmp_conn);
		$axes->send_axes($walk->{'rootSystem'}, $args{'axes_mag'} || 10);
		$axes->xformSet($walk->{'rootSystem'});
	    }
	    $geom = new gvGeom('tilt', $tmp_conn);
	    $geom->xformSet($walk->{'rootSystem'});
	} else {
	    $geom->setConn($tmp_conn);
	}

	# Display the next tilting character
	my $c = $walk->next
	    or die("No character returned\n");
	$geom->send($c);

	$sem->down('empty');
	rename($producer_tmp, "$filename.$next_item")
	    or die("consumer failed to rename $producer_tmp to $filename.$next_item\n");
	$sem->up('full');
	++$next_item >= $buflen and $next_item = 0;
    }
}

# }}}
# {{{ tilt

sub tilt(%) {
    my %args = @_;
    $frames = $args{'frames'} || 100;
    $args{'no_tidy'} and $tidy_up = 0;

    -d $dir
	or mkdir($dir, 0700)
        or die("Failed to create dir $dir\n");
  
    defined (my $pid = fork) or die("Fork failed: $!\n");
  
    if ($pid) {
	# parent consumes and waits
	consumer(%args);
	waitpid ($pid, 0);

    } else {
	# child produces
	producer(%args);
    }
}

# }}}
# {{{ END

sub END {
  # Parent tidies up
  if ($$ == $parent_pid and $tidy_up) {
    unlink <$filename.*>, $producer_tmp, $consumer_tmp;
  }
}

# }}}

1;

__END__

# {{{ Local Variables

# Local Variables:
# mode: perl
# folded-file: t
# End:

# }}}
