# Copyright (C) 2000 Paul Martin and Dave Woodcock. All rights reserved.
# $Id: Vector.pm,v 1.1 2000/11/13 23:59:49 dave Exp $

use strict;
package Vector;

# {{{ new

sub new($@) {
    my ($pkg, @rows) = @_;
    return bless [@rows], $pkg;
}

# }}}
# {{{ minus

# Return a - b
sub minus($$) {
    my ($a, $b) = @_;
    my $sign = 1;
    if (@$a < @$b) {
	($a, $b) = ($b, $a);
	$sign = -1;
    }

    my $c = [ map {$sign * ($a->[$_] - $b->[$_])} 0 .. $#$b ];
    for (my $i = @$b; $i < @$a; ++$i) {
	$c->[$i] = $sign * $a->[$i];
    }
    return bless $c, ref($a);
}

# }}}
# {{{ plus

# Return a + b
sub plus($$) {
    my ($a, $b) = @_;
    ($a, $b) = ($b, $a) if @$a < @$b;

    my $c = [ map {$a->[$_] + $b->[$_]} 0 .. $#$b ];
    for (my $i = @$b; $i < @$a; ++$i) {
	$c->[$i] = $a->[$i];
    }
    return bless $c, ref($a);
}
# }}}
# {{{ div

# Return a / scalar
sub div($$) {
    my ($a, $b) = @_;
    if ($b == 0) {
	use Carp;
	confess ("Cannae divide by zero\n");
    }
    return bless [ map {$a->[$_] / $b} 0 .. $#$a ], ref($a);
}

# }}}
# {{{ mult

# if a is a scalar, returns b scaled by a
# if a is a matrix, return result of operator

sub mult($$) {
    my ($a, $b) = @_;

    if (ref($b) eq 'Matrix') {
	# Matrix case
	my $res;
	for (my $col = 0; $col < @{$b->[0]}; ++$col) {
	    my $s = 0.0;
	    for (my $i = 0; $i < @$a; ++$i) {
		$s += $a->[$i] * $b->[$i]->[$col];
	    }
	    $res->[$col] = $s;
	}
	return bless $res, 'Vector';
	
    } else {
	# Scalar case
	return bless [ map {$a->[$_] * $b} 0 .. $#$a ], ref($a);	
    }
}

# }}}
# {{{ dot

# Dot product of $a and $b
sub dot($$) {
    my ($a, $b) = @_;
    my $s = 0;
    for (my $i = 0; $i < @$a; ++$i) {
	$s += $a->[$i] * $b->[$i];
    }
    return $s;
}

# }}}
# {{{ cross

# Cross product of 2 3-vectors

sub cross($$) {
    my ($a, $b) = @_;
    return bless
	[
	 $a->[1] * $b->[2] - $a->[2] * $b->[1],
	 $a->[2] * $b->[0] - $a->[0] * $b->[2],
	 $a->[0] * $b->[1] - $a->[1] * $b->[0],
	 ], ref($a);
}

# }}}
# {{{ len

# length of a vector
sub len($) {
    my ($a) = @_;
    my $l = 0;
    for (my $i = 0; $i < @$a; ++$i) {
	$l += $a->[$i] * $a->[$i];
    }
    return sqrt($l);
}

# }}}
# {{{ unit

# unit vector parallel to given vector
sub unit($) {
    my ($a) = @_;
    return $a->div($a->len);
}

# }}}
# {{{ print

sub print($) {
    print
	"Vector(",
	join(", ", map {sprintf "%8.4f", $_} @{$_[0]}),
	")\n";
}

# }}}

1;

__END__

# {{{ Local Variables

# Local Variables:
# mode: perl
# folded-file: t
# End:

# }}}
