WARNING;
// Copyright (C) 2000/3 Paul Martin and Dave Woodcock. All rights reserved.
// $Id: exception.m4,v 1.5 2000/11/13 06:16:31 dave Exp $
H_BEGIN(exception);

#include <iostream.h>
// PPM:
#include <stdlib.h>

#ifdef EXCEPTION

// g++ still doesn't seem up to this :-(

// {{{ Exception handling

// {{{ class Exception

class Exception {
  const char* name;
public:
  Exception(const char* m) : name(m) {};
  virtual ostream& print(ostream& s) const {return s << name << ".\n";}
};

class Panic : public Exception {
  const char* msg;
public:
  Panic(const char* n, const char* m) : Exception(n), msg(m) {};
  ostream& print(ostream& s) const {
    s << "Panic: " << msg << " in ";
    return Exception::print(s);
  }
public:
}

inline void PANIC(const char* a, const char* b) {
  throw Panic(a, b);
}

// {{{ class CompatibilityError

class CompatibilityError : public Exception {
public:
  CompatibilityError(const char* m) : Exception(m) {};
};

// {{{ class ArgMismatch

class ArgMismatch : public CompatibilityError {
public:
  ArgMismatch(const char* m)
    : CompatibilityError(m) {};
  ostream& print(ostream& s) {
    s << "Arg mismatch in ";
    return Exception::print(s);
  }
};

inline void ARG_MISMATCH(const char* xxx) {
  throw ArgMismatch(xxx);
}


// }}}
// {{{ class BadArg

class BadArg : public CompatibilityError {
public:
  BadArg(const char* m)
    : CompatibilityError(m) {};
  ostream& print(ostream& s) {
    s << "Bad arg to ";
    return Exception::print(s);
  }
};

inline void BAD_ARG(const char* xxx) {
  throw BadArg(xxx);
}

// }}}
// {{{ class RangeError

class RangeError : public CompatibilityError {
public:
  RangeError(const char* m)
    : CompatibilityError(m) {};
  ostream& print(ostream& s) {
    s << "Range error in ";
    return Exception::print(s);
  }
};

inline void RANGE_ERROR(const char* xxx) {
  throw RangeError(xxx);
}

// }}}

// }}}
// {{{ class AllocationError

class AllocationError : public Exception {
public:
  AllocationError(const char* m) : Exception(m) {};
};

// {{{ class StaticAllocationOverrun

class StaticAllocationOverrun : public AllocationError {
public:
  StaticAllocationOverrun(const char* m)
    : AllocationError(m) {};
  ostream& print(ostream& s) {
    s << "Static allocation exceeded in ";
    return Exception::print(s);
  }
};

inline void STATIC_ALLOCATION_OVERRUN(const char* name) {
  throw StaticAllocationOverrun(name);
}


// }}}

// }}}
// {{{ class DestructionError

class DestructionError : public Exception {
public:
  DestructionError(const char* m) : Exception(m) {};
  
};

// {{{ class BadDeletion

class BadDeletion : public DestructionError {
public:
  BadDeletion(const char *m) : DestructionError(m) {};
  ostream& print(ostream& s) {
    s << "Bad deletion in ";
    return Exception::print(s);
  }
};

inline void BAD_DELETION(const char* xxx) {
  throw BadDeletion(xxx);
}


// }}}

// }}}

// }}}
// {{{ inlines

inline ostream& operator<<(ostream& s, const Exception& e) {
  return e.print(s);
}

// }}}

// }}}

#else // EXCEPTION

// {{{ quit handling

typedef void (*PTF)(); /* pointer to function taking no args
			* and returning no value
			*/

// {{{ class quitObject

class quitObject {
  int num;           // number of currently assignfed exit functions
  int capacity;      // how much space we have for exit functions
  PTF *exitFn;       // array of PTF
  char *errString;   // string to be printed when we quit

public:
  quitObject();
  ~quitObject();
  void add(PTF);
  ostream& print(ostream&) const;
  void setErrString(const char* c) {errString = (char *)c;}
};

// }}}

extern quitObject quit;

// {{{ inline functions

inline ostream& operator<< (ostream& s, const quitObject& q) {
  return q.print(s);
}

inline void operator<< (quitObject& q, const char* c) {
  q.setErrString(c);
}

inline void BAD_ARG(const char* name) {
  cerr << "Bad arg in "      << name << quit;
}
inline void ARG_MISMATCH(const char* name) {
  cerr << "Arg mismatch in " << name << quit;
}
inline void RANGE_ERROR(const char* name) {
  cerr << "Range error in "  << name << quit;
}
inline void BAD_DELETION(const char* name) {
  cerr << "Bad deletion in " << name << quit;
}
inline void PANIC(const char* name, const char* msg) {
  cerr << "Panic: " << msg << " in " << name << quit;
}
inline void STATIC_ALLOCATION_OVERRUN(const char* name) {
  cerr << "Static allocation overrun in " << name << quit;
}

// }}}

#ifdef exception_cc

// {{{ member functions

// {{{ quitObject::print()

ostream& quitObject::print(ostream& s) const {
  // This is the important one.
  // Printing a quit object causes the calling of all
  // specified exit functions, followed by termination
  // of the program.

  if (errString)
    s << "\n" << errString << "\n";
  else
    s << "\n";
  
  for (int i = 0; i < num; ++i)
    (*(exitFn[i]))(); // invoke the function
  exit(1);
  return s;
}

// }}}
// {{{ quitObject::quitObject()

quitObject::quitObject()
  : num(0), capacity(5), errString(0)
{
  exitFn = new PTF[capacity];
}

// }}}
// {{{ quitObject::~quitObject()

quitObject::~quitObject() {
  if (exitFn) delete [] exitFn;
}

// }}}
// {{{ quitObject::add()

void quitObject::add(PTF ptf) {
  if (num >= capacity) {
    capacity = int(1.5 * capacity);
    PTF *tmp = new PTF[capacity];
    for (int i = 0; i < num; ++i)
      tmp[i] = exitFn[i];
    if (exitFn) delete [] exitFn;
    exitFn = tmp;
  }
  exitFn[num++] = ptf;
}

// }}}

// }}}

quitObject quit;

#endif // exception_cc

// }}}

#endif // EXCEPTION

H_END(exception);
C_BEGIN(exception);
C_END(exception);

// {{{ Local Variables

// Local Variables:
// mode: c++
// folded-file: t
// End:

// }}}
