WARNING;
// Copyright (C) 2000 Paul Martin and Dave Woodcock. All rights reserved.
// $Id: misc.m4,v 1.6 2000/11/13 06:16:32 dave Exp $
H_BEGIN(misc);

#include <iostream.h>
#include <string.h>
#include "misc.h"

// {{{ stringCopy

// make a copy of source strings with space allocated by new
char *stringCopy(const char *src1);
char *stringCopy(const char *src1, const char *src2);
char *stringCopy(const char *src1, const char *src2, const char *src3);

// }}}
// {{{ comp

inline int comp(int i, int j) {return i - j;}
inline int comp(float i, float j) {
  if (i > j) return 1;
  else if (i < j) return -1;
  else return 0;
}

// }}}
// {{{ pauseForInt

void pauseForInt();

// }}}
// {{{ dirtyHash

inline unsigned long dirtyHash(unsigned long i) {
  // From a suggestion of Knuth, investigated by H.W. Lewis,
  // described in:
  // Numerical Recipes in C, The Art of Scientific Computation
  // 2nd edition,
  // by Press, Teukolsky, Vetterling and Flannery,
  // C.U.P. (1992).
  return 1664525L * i + 1013904223L;
}

// }}}
// {{{ class string

class string {
  // based heavily on B. Stroustrup's string class in the C++ book

  struct srep {
    char* s; // pointer to data
    int n; // ref count
    srep() : n(1) {};
    srep(int) : s(new char [1]), n(1) {s[0] = 0;} // for initialisation of nullSrep
  };
  
  srep* p;
  static srep* nullSrep;

private:
  string(unsigned int i) : p(new srep) {p->s = new char [i+1];} // just declare the space

public:
  string() : p(nullSrep) {p->n++;};
  string(const char* s) : p(new srep) {
    p->s = new char [strlen(s) + 1];
    strcpy(p->s, s);
  }
  string(const string& s) : p(s.p) {p->n++;}
  ~string() {
    if (--p->n == 0) {
      delete [] p->s;
      delete p;
    }
  }
  unsigned int len() const {return strlen(p->s);}
  string& operator= (const char*);
  string& operator= (const string&);
  operator char* () {return p->s;}
  operator const char* () const {return p->s;}
  bool operator== (const string& s) const {return strcmp(p->s, s.p->s) == 0;}
  bool operator!= (const string& s) const {return strcmp(p->s, s.p->s) != 0;}
  bool operator<  (const string& s) const {return strcmp(p->s, s.p->s) < 0;}
  bool operator<= (const string& s) const {return strcmp(p->s, s.p->s) <= 0;}
  bool operator>  (const string& s) const {return strcmp(p->s, s.p->s) > 0;}
  bool operator>= (const string& s) const {return strcmp(p->s, s.p->s) >= 0;}
  char& operator[] (unsigned int i) {return p->s[i];}
  string operator+(const string&) const;
  ostream& print(ostream& x) const {return x << p->s;}
  unsigned long hash() const {
    unsigned long seed = 1030501; // DJW! Is this random?
    for(char* s = p->s; *s;)
  	 seed = dirtyHash(seed ^ *s++);
    return seed;
  }
};

// }}}
// {{{ inlines

inline ostream& operator<<(ostream& s, const string& x) {
  return x.print(s);
}

inline unsigned long hash(const string& s) {
  return s.hash();
}

// }}}

H_END(misc);
C_BEGIN(misc);

#include <string.h>

// {{{ stringCopy()

char *stringCopy(const char *src1) {
  char *dest = new char[strlen(src1) + 1];
  strcpy(dest, src1);
  return dest;
}

char *stringCopy(const char *src1, const char *src2) {
  char *dest = new char[strlen(src1) + strlen(src2) + 1];
  strcpy(dest, src1);
  strcat(dest, src2);
  return dest;
}

char *stringCopy(const char *src1, const char *src2, const char *src3) {
  char *dest = new char[strlen(src1) + strlen(src2) + strlen(src3) + 1];
  strcpy(dest, src1);
  strcat(dest, src2);
  strcat(dest, src3);
  return dest;
}

// }}}
// {{{ pauseForInt()

void pauseForInt() {
  cout << "Enter a number to continue\n";
  int c;
  cin >> c;
}

// }}}
// {{{ static stuff

string::srep* string::nullSrep = new string::srep(0);

// }}}
// {{{ string members

// {{{ string::operator=()

string& string::operator=(const char* s) {
  if (p->n > 1) { // disconnect self
    p->n--;
    p = new srep;
  } else {
    delete [] p->s; // free old space
  }
  p->s = new char [strlen(s) + 1];
  strcpy(p->s, s);
  return * this;
}

string& string::operator=(const string& s) {
  s.p->n++; // protect against this == s
  if (--p->n == 0) {
    delete [] p->s;
    delete p;
  }
  p = s.p;
  return *this;
}

// }}}
// {{{ string::operator+()

string string::operator+(const string& s) const {
  unsigned int l = len();
  string res(l + s.len());
  strcpy(res.p->s, p->s);
  strcpy(res.p->s + l, s.p->s);
  return res;
}

// }}}

// }}}

C_END(misc);

// {{{ Local Variables

// Local Variables:
// mode: c++
// folded-file: t
// End:

// }}}
