WARNING;
// Copyright (C) 2000 Paul Martin and Dave Woodcock. All rights reserved.
// $Id: urhash.m4,v 1.6 2000/11/13 06:16:33 dave Exp $
H_BEGIN(urhash);

#include <iostream.h>
#include "exception.h"

// Parameterised by classes X (object stored) and K (the key)
// Requires the following public members:
//
//    unsigned long hash(const K&);
//    bool X::operator==(const K&);
//    X(const K&);
//    X* X::next();
//    X* X::setNext(X*);
//    ostream& print(ostream&);

// {{{ class urhash

template <class X, class K>
class urhash {

private:
  const X** box;
  unsigned long mask;
  unsigned int reportRate;
  unsigned int numEntries;
  unsigned int numAccesses;
  unsigned int numCollisions;

  // disallow these
  urhash(const urhash<X,K>&);
  urhash<X,K>& operator=(const urhash<X,K>&);
		      
public:			
  urhash();
  urhash(unsigned int numBits, unsigned int reprate);
  // void init(unsigned int numBits, unsigned int reprate);
  ~urhash();
  const X* addr(const K&, bool nocreate=false);
  ostream& print(ostream&) const;
  // void report();
};

// }}}
// {{{ inline members

// {{{ urhash<X,K>::addr()

template<class X, class K>
inline const X* urhash<X,K>::addr(const K& k, bool nocreate) {
  unsigned long h = hash(k) & mask;
  const X* present = box[h];
  const X* previous = 0;

  while (present) {
    if (*present == k) return present;
    previous = present;
    present = present->next();
   }

  // didn't find it
  if (nocreate) {
    return 0;
  } else {
    // create new entry
    present = new X(k);
    return previous ? previous->setNext(present) : (box[h] = present);
  }
}

// template<class X, class K>
// inline X* urhash<X,K>::addr(const X& x) {
//   unsigned int h = x.hash() & mask;
//   X* ptr = box[h];
//   X* prev = 0;
// 
//   while (ptr) {
//     if (*ptr == x) return ptr;
//     prev = ptr;
//     ptr = ptr->next();
//   }
// 
//   // didn't find it so create new entry
//   ptr = new X(x);
//   return prev ? prev->setNext(ptr) : (box[h] = ptr);
// }

// template<class X, class K>
// X* urhash<X,K>::addr(const X& x) {
//   unsigned int h = x.hash() & mask;
//   bool collision = false;
//   X* ptr = box[h];
//   X* prev = 0;
// 
//   if ((++numAccesses % reportRate) == 0) {
//     report();
//   }
//   
//   while (ptr) {
//     if (*ptr == x) {
// 	 if (collision) ++numCollisions;
// 	 return ptr;
//     }
//     prev = ptr;
//     ptr = ptr->next();
//     collision = true;
//   }
// 
//   // didn't find it
//   ptr = new X(x);
//   if (prev) {
//     return prev->setNext(ptr);
//   } else {
//     ++numEntries;
//     return box[h] = ptr;
//   }
// }

// }}}

// }}}
// {{{ members

// {{{ urhash<X,K>::print()

template<class X, class K>
ostream& urhash<X,K>::print(ostream& s) const {
  for (unsigned long i = 0; i <= mask; ++i) {
    const X* present = box[i];
    while (present) {
      present->print(s);
      s << " ";
      present = present->next();
    }
  }
  return s;
}

// }}}
// {{{ urhash<X,K>::urhash()

// {{{ ()

template<class X,class K>
urhash<X,K>::urhash()
  :
  mask(0),
  reportRate(0),
  numEntries(0),
  numAccesses(0),
  numCollisions(0)
{}

// }}}
// {{{ (unsigned int, unsigned int)

template<class X, class K>
urhash<X,K>::urhash(unsigned int numBits, unsigned int reprate)
  :
  mask((1 << numBits) - 1),
  reportRate(reprate),
  numEntries(0),
  numAccesses(0),
  numCollisions(0)
{
  box = (const X**) new X* [mask + 1];
  for (unsigned int i = mask + 1; i--;)
    box[i] = 0;
}

// }}}

// }}}
// {{{ urhash<X,K>::~urhash()

template<class X, class K>
urhash<X,K>::~urhash() {
  if (box) delete [] box;
}

// }}}
// {{{ urhash<X,K>::init() (commented out)

// template<class X, class K>
// void urhash<X,K>::init(unsigned int numBits, unsigned int reprate) {
//   if (mask)
//     PANIC("urhash<X,K>::init()",
// 	     "already initialised.");
//   mask = (1 << numBits) - 1;
//   reportRate = reprate;
//   numEntries = 0;
//   numAccesses = 0;
//   numCollisions = 0;
//   box = (const X**) new X* [mask + 1];
//   for (unsigned int i = mask; i--;)
//     box[i] = 0;
// }

// }}}
// {{{ urhash<X,K>::report()

// template<class X, class K>
// void urhash<X,K>::report() {
//   cout << "entr: " << numEntries << " "
// 	  << "col: "
// 	  << (float(numCollisions * 100) / float(numAccesses)) << "% "
// 	  << "cap: "
// 	  << (float(numEntries * 100) / float(mask + 1)) << "%\n";
//   cout.flush();
// }

// }}}

// }}}

H_END(urhash);

// {{{ Local Variables

// Local Variables:
// mode: c++
// folded-file: t
// End:

// }}}
