WARNING;
// Copyright (C) 2000 Paul Martin and Dave Woodcock. All rights reserved.
// $Id: affineWeyl.m4,v 1.10 2000/11/13 06:16:35 dave Exp $
H_BEGIN(affineWeyl);

#include <iostream.h>
#ifdef EXTRA_SAFE
#  include "exception.h"
#endif
#include "intvec.h"
#include "intmat.h"
#include "perm.h"
#include "rootSystem.h"

class rootSystem;
class alcove;

// {{{ class affineWeyl

class affineWeyl {
private:
  int idx;
  intvec trans;

public:
  rootSystem* R;

public:
  affineWeyl() : idx(0), trans(), R(0) {}; // needed for building arrays
  affineWeyl(rootSystem* r, unsigned int i, const intvec& t);
  affineWeyl(const affineWeyl&);
  ~affineWeyl();
  affineWeyl& operator=(const affineWeyl&);
  affineWeyl operator*(const affineWeyl&) const;
  intvec rightMultiple(const intvec&) const;
  affineWeyl& operator*=(const affineWeyl&);
  affineWeyl inv() const;
  bool operator==(const affineWeyl&) const;
  bool operator!=(const affineWeyl&) const;
  intvec centre() const;
  const alcove * alc() const; // the associated alcove
  unsigned long hash() const;
  ostream& print(ostream&) const;
};

// }}}
// {{{ inline functions

// {{{ affineWeyl::affineWeyl()

inline affineWeyl::affineWeyl(rootSystem* r, unsigned int i, const intvec& t) 
  : idx(i), trans(t), R(r) {
#ifdef EXTRA_SAFE
    if ((R == 0) || (i >= R->weylSize))
      BAD_ARG("affineWeyl::affineWeyl()");
#endif
}

// }}}
// {{{ affineWeyl::affineWeyl()

inline affineWeyl::affineWeyl(const affineWeyl& w) 
  : idx(w.idx), trans(w.trans), R(w.R) {
}

// }}}
// {{{ operator<< ()

inline ostream& operator<<(ostream&s, const affineWeyl& w) {
  return w.print(s);
}

// }}}
// {{{ affineWeyl::~affineWeyl()

inline affineWeyl::~affineWeyl() {
}

// }}}
// {{{ affineWeyl::hash()

inline unsigned long affineWeyl::hash() const {
  return ::hash(trans, idx);
}

inline unsigned long hash(const affineWeyl& w) {
  return w.hash();
}

// }}}
// {{{ operator*(const intvec&, const affineWeyl&)

inline intvec operator*(const intvec&v, const affineWeyl& w) {
  return w.rightMultiple(v);
}

// }}}

// }}}

H_END(affineWeyl);
C_BEGIN(affineWeyl);

#include <stdlib.h>
#include "alcove.h"
#include "alcovehash.h"
#include "rootSystem.h"

// {{{ nonstatic members

// {{{ affineWeyl::operator=()

affineWeyl& affineWeyl::operator=(const affineWeyl& w) {
  R = w.R;
  idx = w.idx;
  trans = w.trans;
  return *this;
}

// }}}
// {{{ affineWeyl::operator*()

affineWeyl affineWeyl::operator*(const affineWeyl& w) const {

#ifdef EXTRA_SAFE
  if (R != w.R)
    ARG_MISMATCH("affineWeyl::operator*():\n");
#endif

  return affineWeyl(R,
		    idx * R->weylPerm[w.idx],
		    (trans * R->weylMat[w.idx]) + w.trans);
}

// }}}
// {{{ affineWeyl::operator*=()

affineWeyl& affineWeyl::operator*=(const affineWeyl& w) {

#ifdef EXTRA_SAFE
  if (R != w.R)
    ARG_MISMATCH("affineWeyl::operator*=():\n");
#endif

  idx = idx * R->weylPerm[w.idx];
  trans = (trans * R->weylMat[w.idx]) + w.trans;
  return *this;
}

// }}}
// {{{ affineWeyl::inv()

affineWeyl affineWeyl::inv() const {
  int j = R->weylInv[idx];
  return affineWeyl(R, j, -trans * R->weylMat[j]);
}

// }}}
// {{{ affineWeyl::operator==()

bool affineWeyl::operator==(const affineWeyl& w) const {
  return (idx == w.idx) && (trans == w.trans) && (R == w.R);
}

// }}}
// {{{ affineWeyl::operator!=()

bool affineWeyl::operator!=(const affineWeyl& w) const {
#ifdef EXTRA_SAFE
  if (R != w.R)
    ARG_MISMATCH("affineWeyl::operator!=()");
#endif

  return (idx != w.idx) || (trans != w.trans);
}

// }}}
// {{{ affineWeyl::centre()

intvec affineWeyl::centre() const {
  return intvec(R->alcoveCentre[idx] + trans);
}

// }}}
// {{{ affineWeyl::alcove

const alcove* affineWeyl::alc() const {
  // DJW! All this casting is because of definition loops:
  // rootSystem needs alcovehash needs alcove needs affineWeyl needs rootSystem
  return (const alcove*)(((alcovehash*)(this->R->alchash))->addr(*this));
}

// }}}
// {{{ affineWeyl::rightMultiple()

intvec affineWeyl::rightMultiple(const intvec& v) const {
#ifdef EXTRA_SAFE
  if (v.len() != R->rank)
    ARG_MISMATCH("affineWeyl::rightMultiple()");
#endif

  return intvec(v * R->weylMat[idx] + trans);
}

// }}}
// {{{ affineWeyl::print()

ostream& affineWeyl::print(ostream &s) const {
  return s << "[" << idx << "," << trans << "]";
}

// }}}

// }}}

C_END(affineWeyl);

// {{{ Local Variables

// Local Variables:
// mode: c++
// folded-file: t
// End:

// }}}
