WARNING;
// Copyright (C) 2000 Paul Martin and Dave Woodcock. All rights reserved.
// $Id: alcove.m4,v 1.11 2000/11/13 06:16:36 dave Exp $
H_BEGIN(alcove);

// Alcoves should in practice always be accessed through pointers.
// It makes no sense to have two copies of the same alcove.

#include "exception.h"
#include "intvec.h"
#include "intmat.h"
#include "intpol.h"
#include "rootSystem.h"
#include "affineWeyl.h"

// {{{ class alcove

class character;
class alcove;
typedef const alcove Alcove;

class alcove {
private:
  affineWeyl  ww;             // corresponding affine Weyl group elt
  bool        iisDominant;    // are we dominant?
  bool        *rreflectsDown; // do we reflect down?
  Alcove**    rreflection;    // our neighbours
  Alcove*     nnext;          // for hash collisions
  character*  ttiltChar;      // tilting character based at this alcove

private:
  // Should never copy an alcove, so make the
  // copy constructor and operator private.
  // I'd like to do the same with the destructor, but the compiler whines.

  alcove(alcove&);
  alcove operator=(alcove&);


private:
  // tmpPol is used during character computations to avoid
  // having to sort lists of alcoves, and during gvPrintExteriorWalls
  // to figure out which walls are on the outside.
  //
  // IT MUST BE NULL EXCEPT DURING SUCH A COMPUTATION,
  // APART FROM ITS INITIALISATION AS ZERO.
  // CONSEQUENTLY, libtilt IS NOT THREAD SAFE! 

  intpol* ttmpPol;

public:
  alcove(const affineWeyl&);
  ~alcove() {
    BAD_DELETION("alcove::~alcove()");
  }

public:
  // properties
  bool isDominant() const {return iisDominant;}
  bool reflectsDown(int) const;
  bool operator==(const affineWeyl& w) const {return ww == w;}

public:
  // stuff deferred to ww and its root system
  const affineWeyl& w() const {return ww;}
  intvec centre() const {return ww.centre();}
  const unsigned int rank() const {return ww.R->rank;}
  const intmat& cartanMat() const {return ww.R->cartanMat;}
  const affineWeyl& s(unsigned int i) const {return ww.R->s[i];}
  const intvec& alcoveFace(unsigned int i) const {return ww.R->alcoveFace[i];}
  const intvec& fundamentalVertex(unsigned int i) const {return ww.R->fundamentalVertex[i];}

public:
  Alcove* reflection(unsigned int) const;
  Alcove* next() const {return nnext;}
  Alcove* setNext(Alcove *a) const {return ((alcove *)(this))->nnext = a;}
  ostream& print(ostream&) const;
  void gvPrint(ostream*) const;
  void gvPrintWall(ostream*, unsigned int) const;
  const character* tiltChar() const {return ttiltChar;}
  void setTiltChar(character* c) const {((alcove *)(this))->ttiltChar = c;}
  intpol*& tmpPol() const {return ((alcove *)(this))->ttmpPol;}
};

// }}}
// {{{ inline functions

inline ostream& operator<<(ostream& s, Alcove& a) {
  return a.print(s);
}

// }}}

H_END(alcove);
C_BEGIN(alcove);

#include <stdlib.h>
#include "gvObj.h"
#include "misc.h"
#include "alcovehash.h"
#include "affineWeyl.h"

// {{{ nonstatic members

// {{{ alcove::alcove()

alcove::alcove(const affineWeyl& w)
  : ww(w),
    iisDominant(true),
    nnext(0),
    ttiltChar(0),
    ttmpPol(0)

{
  // Needs coercing
  rreflectsDown = new bool  [rank()+1];
  rreflection   = new Alcove* [rank()+1];

  for (int i = rank(); i >= 0; --i)
    rreflection[i] = 0;

  intvec v(centre() * cartanMat());
  for (int i = rank(); i--;)
    if (v[i] < 0) {
      iisDominant = false;
      break;
    }
}

// }}}
// {{{ alcove::reflection()

Alcove* alcove::reflection(unsigned int i) const {
  // This function is only logically const,
  // so we coerce to non-const via `This'.

  if (!rreflection[i]) {
    // haven't yet computed it

    // Coerce to nonconst
    alcove* This = (alcove *)this;
    affineWeyl r = s(i) * ww;
    This->rreflection[i] = r.alc();

    // Compute relative positions
    const intvec ccentre  = centre();
    const intvec rcentre  = rreflection[i]->centre();

    if (ccentre > rcentre)
      This->rreflectsDown[i]
	= (rcentre * cartanMat() * (ccentre - rcentre)) > 0;
    else
      This->rreflectsDown[i]
	= (ccentre * cartanMat() * (rcentre - ccentre)) < 0;
  }

  return rreflection[i];
}

// }}}
// {{{ alcove::reflectsDown()

bool alcove::reflectsDown(int i) const {
  if (!rreflection[i])
    // Haven't computed the reflection,
    // so do it now:
    reflection(i);

  return rreflectsDown[i];
}

// }}}
// {{{ alcove::print()

ostream& alcove::print(ostream& s) const {
  return s << "A" << centre();
}

// }}}
// {{{ alcove::gvPrintWall()

void alcove::gvPrintWall(ostream* s, unsigned int j) const {
  if (rank() != 3)
    BAD_ARG("alcove::gvPrintWall()");

  *s << "OFF 3 1 0\n"; // number of vertices, faces and edges (latter ignored)
  const intvec& v = alcoveFace(j);
  // the points
  for (unsigned int i = 0; i < 3; ++i) {
    intvec u = fundamentalVertex(v[i]) * ww;
    *s << u[0] << " " << u[1] << " " << u[2] << "\n";
  }
  *s << "3 0 1 2 " // number of vertices in the face, followed by their indices
     << colour::Colour[j] << "\n";
}

// }}}
// {{{ alcove::gvPrint()

void alcove::gvPrint(ostream* s) const {
  static intvec u;

  switch(rank()) {
  case 2:
    *s << "OFF 3 1 0\n"; // number of vertices, faces and edges (latter ignored)
    // the points
    for (int i = 0; i < 3; ++i) {
      u = fundamentalVertex(i) * ww;
      *s << u[0] << " " << u[1] << " 0\n";
    }
    // the face
    *s << "3 0 1 2\n"; // number of vertices in the face, followed by their indices
    return;

  case 3:
    *s << "OFF 4 4 0\n"; // number of vertices, faces and edges (latter ignored)
    // the points
    for (int i = 0; i < 4; ++i) {
      u = fundamentalVertex(i) * ww;
      *s << u[0] << " " << u[1] << " " << u[2] << "\n";
    }
    // the faces
    for (int i = 0; i < 4; ++ i) {
      u = alcoveFace(i);
      *s << "3 " << u[0] << " " << u[1] << " " << u[2] << " "
	 << colour::Colour[i] << "\n";
    }
    return;

  default:
    BAD_ARG("alcove::gvPrintAlcove()");
  }
}

// }}}

// }}}

C_END(alcove);

// {{{ Local Variables

// Local Variables:
// mode: c++
// folded-file: t
// End:

// }}}
