WARNING;
// Copyright (C) 2000 Paul Martin and Dave Woodcock. All rights reserved.
// $Id: character.m4,v 1.13 2000/11/14 23:20:14 dave Exp $
H_BEGIN(character);

#include "intpol.h"
#include "intpolhash.h"

class alcove;
typedef const alcove Alcove;

// {{{ class character

class character {
  unsigned int num;
  Alcove** aa;
  Intpol** pp;

  static unsigned int buflen;
  static Alcove** aBuf;
  static intpol* pBuf; // CAUTION: this is intpol* while pp is intpol** 

  static void checkBuf(unsigned int); // get enough buffer space

public:

  character() : num(0), aa(0), pp(0) {};

  character(Alcove* al, Intpol* pol) : num(1) {
    aa = (Alcove **)(new alcove* [1]);
    pp = (Intpol **)(new intpol* [1]);
    aa[0] = al;
    pp[0] = pol;
  }

  character(character& c) : num(c.num) {
    aa = (Alcove **)(new alcove* [num]);
    pp = (Intpol **)(new intpol* [num]);
    for(unsigned int i = 0; i < num; ++i) {
      aa[i] = c.aa[i];
      pp[i] = c.pp[i];
    }
  };

  ~character() {
    if (aa) delete [] aa;
    if (pp) delete [] pp;
  };

  Alcove* a(unsigned int i) const {return aa[i];}
  Intpol& p(unsigned int i) const {return *(pp[i]);}
  unsigned int len() const {return num;}
  character* reflection(unsigned int) const; // creates a new character
  void cancel(const character* c, int i);    // cancels i times c
  void cullZeroes();
  void gvPrintExteriorWalls(ostream*) const;
  ostream& print(ostream&) const;
};

// }}}
// {{{ inline stuff

inline ostream& operator<< (ostream& s, const character& c) {
  return c.print(s);
}

// }}}

H_END(character);
C_BEGIN(character);

#include <stdlib.h>
#include "alcove.h"
#include "exception.h"

// {{{ character member functions

unsigned int character::buflen = 0;
Alcove** character::aBuf = 0;
intpol* character::pBuf = 0;

// {{{ character::checkBuf(int l)

void character::checkBuf(unsigned int l) {
  if (l > buflen) {
    unsigned int oldBuflen = buflen;

    // Check to see the entries of pBuf are ok:
    // cout << "Old entries of pBuf:\n";
    // for (int i = 0; i < buflen; ++i)
    // 	 cout << "  " << pBuf[i] << "\n";

    buflen = (unsigned int)(buflen * 1.5);
    if (buflen < l) buflen = l;
    if (oldBuflen) {
      delete [] aBuf;
      delete [] pBuf;
    }
    aBuf = (Alcove **)(new alcove*[buflen]);
    pBuf = new intpol[buflen];
  }
}

// }}}
// {{{ character::reflection()

character* character::reflection(unsigned int i) const {
  checkBuf(2 * num);
  unsigned int j, k;
  
  for (j = k = 0; j < num; ++j) {
    bool down = aa[j]->reflectsDown(i);
    Alcove *refl = aa[j]->reflection(i);

    // !!! does this test save me any time?
    if (down || refl->isDominant()) {
      if (aa[j]->tmpPol()) {
	// Alcove already encountered in this computation
	*(aa[j]->tmpPol()) += (*(pp[j]))[down ? -1 : 1];
      } else {
	// Haven't seen it yet
	aBuf[k] = aa[j];
	aa[j]->tmpPol() = pBuf + k;
	pBuf[k++] = (*(pp[j]))[down ? -1 : 1];
      }      
    
      if (refl->tmpPol()) {
	// Alcove already encountered in this computation
	*(refl->tmpPol()) += *(pp[j]);
      } else {
	// Haven't seen it yet
	aBuf[k] = refl;
	refl->tmpPol() = pBuf + k;
	pBuf[k++] = *(pp[j]);
      }
    }
  }
  
  character* result = new character;
  result->num = k;
  result->aa = (Alcove **)(new alcove* [k]);
  result->pp = (Intpol **)(new intpol* [k]);
  for (j = 0; j < k; ++j) {
    result->aa[j] = aBuf[j];
    result->pp[j] = polhash.addr(pBuf[j]);
    // Zero the tmpPol's we've used
    aBuf[j]->tmpPol() = 0;
  }
  return result;
}

// }}}
// {{{ character::cancel()

void character::cancel(const character* c, int i) {
  // cancel i copies of c
  // Note that we do NOT remove alcoves with zero coefficient,
  // because of the way this functions is used in alcove::tiltChar()

  unsigned int j;
  
  // Flag the alcoves we currently use
  checkBuf(num);
  
  for (j = 0; j < num; ++j) {
    aa[j]->tmpPol() = pBuf +j;
    pBuf[j] = *(pp[j]);
  }

  // Perform the subtractions
  for (j = 0; j < c->num; ++j)
    if (c->aa[j]->tmpPol())
      c->aa[j]->tmpPol()->addMultiple(*(c->pp[j]), -i);
    else
      BAD_ARG("character::cancel()");
  
  for (j = 0; j < num; ++j) {
    if (pBuf[j] != *(pp[j]))
      // It's changed, so need to recompute its address
      pp[j] = polhash.addr(pBuf[j]);
    aa[j]->tmpPol() = 0;
  }
}

// }}}
// {{{ character::cullZeroes()

void character::cullZeroes() {
  static unsigned int i,j;
  for (i = j = num; i--;)
    if (pp[i]->isZero()) --j;

  if (j < num) {
    // some space needs reclaiming
    Alcove** aa1 = aa;
    Intpol** pp1 = pp;
    aa = (Alcove **)(new alcove* [j]);
    pp = (Intpol **)(new intpol* [j]);
    for (i = num, num = j; i--;)
      if (!pp1[i]->isZero()) {
	aa[--j] = aa1[i];
	pp[j]   = pp1[i];
      }
    delete [] aa1;
    delete [] pp1;
  }
}

// }}}
// {{{ character::print()

ostream& character::print(ostream& s) const {
  s << "C[";
  for (unsigned int i = 0; i < num; ++i)
    if (i == 0)
      s << "(" << *(pp[i]) << ") * " << *aa[i];
    else
      s << ", (" << *(pp[i]) << ") * " << *aa[i];
  return s << "]";
}

// }}}
// {{{ character::gvPrintExteriorWalls()

void character::gvPrintExteriorWalls(ostream* s) const {
  // print an OOGL spec corresponding to the exterior walls
  // of the support of the character

  if (len() == 0) return;
  unsigned int rank = aa[0]->rank();

  switch(rank) {
  case 2:
    *s << "LIST\n";
    for (unsigned int i = 0; i < len(); ++i)
      aa[i]->gvPrint(s);
    *s << "\n";
    return;
    
  case 3:
    *s << "LIST\n";

    // flag the alcoves in the support
    for (unsigned int i = 0; i < len(); ++i)
      aa[i]->tmpPol() = (intpol*)(pp[0]);
  
    for (unsigned int i = 0; i < len(); ++i)
      for (unsigned int j = 0; j <= rank; ++j)
	if (aa[i]->reflection(j)->tmpPol() == 0)
	  aa[i]->gvPrintWall(s, j);

    // clear the alcoves in the support
    for(unsigned int i = 0; i < len(); ++i)
      aa[i]->tmpPol() = 0;

    *s << "\n";
    return;

  default:
    BAD_ARG("character::gvPrintExteriorWalls()");
  }
}

// }}}

// }}}

C_END(character);

// {{{ Local Variables

// Local Variables:
// mode: c++
// folded-file: t
// End:

// }}}
