WARNING;
// Copyright (C) 2000 Paul Martin and Dave Woodcock. All rights reserved.
// $Id: rootSystem.m4,v 1.13 2000/11/14 00:03:16 dave Exp $
H_BEGIN(rootSystem);

#include "intvec.h"
#include "intmat.h"
#include "floatmat.h"

class perm;
class affineWeyl;

// {{{ class rootSystem

class rootSystem {
  // We work in the free abelian group on the simple roots.
  // In some places it is necessary to introduce a scaling to
  // make some rational vectors (e.g. the fundamental weights)
  // integral.

  // private:
  intmat * weylMat;          // matrices for ordinary Weyl group elts
  intvec weylLength;         // lengths of ordinary Weyl group elts
  perm * weylPerm;           // regular permutation rep of Weyl
  intvec weylInv;            // lookup table for inverses
  intvec rootReflection;     /* lookup for the reflections, indexed
			      * by the positive root labels */
  // intvec simpleRootLength;
  intvec *simpleRoot;
  intvec *positiveRoot;
  unsigned int    highestShortRoot;   // index of highest short root among pos roots
  intvec *fundamentalWeight; /* these are scaled to be integral vectors
				in the simple root basis */
  intvec *alcoveCentre;      // alcove centres of ordinary Weyl group
  intvec *canonicalCentre;

  void release();
  void setup(const char* CM);
  void buildCartanMat();
  void buildWeylMat();
  void buildRoots();
  void buildWeylPerm();
  void buildAlcoves();
  void buildCoxeterGenerators();

public:
  unsigned int rank;         // the `n' in An etc
  char type;    	     // A,B,C,D,E,F or G
  char * ident; 	     // "A3" etc
  unsigned int prime;        // l or p; it determines alcove size
  unsigned int numPositiveRoots;
  unsigned int weylSize;     // number of elts in ordinary Weyl group
  void* alchash;             // look up of alcoves by affine Weyl elt (it's an alcovehash)
  intmat cartanMat;
  intmat bilinearForm;       // scaled by a factor of 2
  intmat identityMat;
  intvec *fundamentalVertex; // vertices of fundamental alcove (scaled)
  intvec *alcoveFace;        /* lists indices of the points of the faces of the
				fundamental alcove */

  affineWeyl *s;             // Coxeter generators of affine Weyl group
  floatmat onTransform;      // transform to orthonormal basis
  intvec rho;                // sum of scaled fundamental weights  
  intvec origin;             // the zero vector

  rootSystem(const char* CM);
  ~rootSystem() {release();}
  void reset(const char* CM);
  ostream& print(ostream&) const;
  ostream& printWeylGroup(ostream&) const;
  affineWeyl aw(const intvec& v) const;  // affine Weyl elt corresponding to
                                         // a sequence of indices
  void gvPrintAxes(ostream*, int) const;

  // Not approved, but it seems tidiest
  friend class affineWeyl;
};

// }}}
// {{{ inline functions

inline ostream& operator<<(ostream& s, const rootSystem& r) {
  return r.print(s);
}

// }}}

H_END(rootSystem);
C_BEGIN(rootSystem);

#include <stdlib.h>
#include "misc.h"
#include "hcf.h"
#include "floatvec.h"
#include "perm.h"
#include "affineWeyl.h"
#include "exception.h"
#include "alcovehash.h"

// {{{ member functions

// {{{ rootSystem::release()

void rootSystem::release() {
  if (ident) delete [] ident;
  if (weylMat) delete [] weylMat;
  if (weylPerm) delete [] weylPerm;
  if (simpleRoot) delete [] simpleRoot;
  if (positiveRoot) delete [] positiveRoot;
  if (fundamentalWeight) delete [] fundamentalWeight;
  if (fundamentalVertex) delete [] fundamentalVertex;
  if (s) delete [] s;
  if (alcoveCentre) delete [] alcoveCentre;
  if (canonicalCentre) delete [] canonicalCentre;
  if (alcoveFace) delete [] alcoveFace;
  if (alchash) delete alchash;
}

// }}}
// {{{ rootSystem::setup()

void rootSystem::setup(const char* CM) {
  ident = stringCopy(CM);
  type = ident[0];
  rank = atoi(ident + 1);
  
  buildCartanMat(); // also builds bilinearForm and simpleRootLength
  
  identityMat = intmat::idMat(rank);
   
  // simple roots and fundamental weights
  simpleRoot = new intvec[rank];
  fundamentalWeight = new intvec[rank];
  intmat invCartanMat = quasiInv(cartanMat);
  rho = origin = intvec(rank, 0);
  
  for (unsigned int i = 0; i < rank; ++i) {
    simpleRoot[i] = identityMat.row(i);
    fundamentalWeight[i] = invCartanMat.row(i);
    rho += fundamentalWeight[i];
  }
  
  buildWeylMat();
  buildWeylPerm();
  buildRoots();
  buildAlcoves();
  buildCoxeterGenerators();
  
  onTransform = Floatmat(identityMat);
  GramSchmidt(onTransform, Floatmat(bilinearForm));
  onTransform = inv(onTransform);
}

// }}}
// {{{ rootSystem::rootSystem()

rootSystem::rootSystem(const char* CM)
  :
  weylMat(0),
  weylPerm(0),
  simpleRoot(0),
  positiveRoot(0),
  fundamentalWeight(0),
  alcoveCentre(0),
  canonicalCentre(0),
  ident(0),
  alchash(new alcovehash(14, 0)), // 16K entries
  fundamentalVertex(0),
  alcoveFace(0),
  s(0)
{
  // CM is the name of an indecomposable
  //  Cartan matrix in standard form, e.g. "A4".
  setup(CM);
}

// }}}
// {{{ rootSystem::reset()

void rootSystem::reset(const char* CM) {
  release();
  setup(CM);
}

// }}}

// {{{ rootSystem::print()

ostream& rootSystem::print(ostream& s) const {
  s << "Root system:\n"
    << "Type:          " << ident << "\n"
    << "prime:         " << prime << "\n"
    << "Cartan matrix: " << cartanMat << "\n"
    << "num pos roots: " << numPositiveRoots << "\n"
    << "id mx:         " << identityMat << "\n"
    << "origin:        " << origin << "\n"
    << "rho:           " << rho << "\n";

  s << "fundamental weights:\n";
  for (unsigned int i = 0; i < rank; ++i)
    s << fundamentalWeight[i] << "\n";

  s << "Weyl size:     " << weylSize << "\n"
    << "lengths of Weyl group elts: " << weylLength << "\n";

  s << "Weyl permutations:\n";
  for (unsigned int i = 0; i < weylSize; ++i)
    s << weylPerm[i] << "\n";
  
  s << "Positive roots and their root reflections:\n";
  for (unsigned int i = 0; i < numPositiveRoots; ++i)
    s << positiveRoot[i] << ", " << rootReflection[i] << "\n";
  
  s << "highest short root:" << positiveRoot[highestShortRoot] << "\n"
    << "vertices of fundamental alcove:\n";
  for (unsigned int i = 0; i <= rank; ++i)
    s << fundamentalVertex[i] << "\n";
  s << "centre of fundamental alcove: " << alcoveCentre[0] << "\n"
    << "on transform: " << onTransform << "\n";

  return s;
}

// }}}
// {{{ rootSystem::printWeylGroup()

ostream& rootSystem::printWeylGroup(ostream& s) const {
  // Print ordinary Weyl group as an OOGL TLIST
  // This is handy for objects with Weyl symmetry:
  // one only has to give data for a Weyl orbit transversal.

  switch(rank) {
  case 3:
    s << "TLIST\n";
    for (unsigned int i = 0; i < weylSize; ++i) {
      intmat& m = weylMat[i];
      s << m(0,0) << " " << m(0,1) << " " << m(0,2) << " 0\n"
	<< m(1,0) << " " << m(1,1) << " " << m(1,2) << " 0\n"
	<< m(2,0) << " " << m(2,1) << " " << m(2,2) << " 0\n"
	<< "0 0 0 1\n";
    }
    return s;

  case 2:
    s << "TLIST\n";
    for (unsigned int i = 0; i < weylSize; ++i) {
      intmat& m = weylMat[i];
      s << m(0,0) << " " << m(0,1) << " 0 0\n"
	<< m(1,0) << " " << m(1,1) << " 0 0\n"
	<< "0 0 1 0\n"
	<< "0 0 0 1\n";
    }
    return s;

  default:
    BAD_ARG("rootSystem::printWeylGroup()");
  }
  // can't get here, but gcc doesn't know that
  return s;
}

// }}}
// {{{ rootSystem::aw()

affineWeyl rootSystem::aw(const intvec& v) const {
  // affine Weyl elt corresponding to
  // a sequence of indices

  affineWeyl result((rootSystem *)this, 0, origin);
  unsigned int n = v.len();
  int j;

  for (unsigned int i = 0; i < n; ++i)
    if (((j = v[i]) >= 0) && (j <= int(rank))) // cast for the compiler's benefit
      result *= s[v[i]];
    else
      BAD_ARG("rootSystem::aw()");
  return result;
}

// }}}
// {{{ rootSystem::buildCartanMat()

void rootSystem::buildCartanMat() {
  static char *fnName = "buildCartanMat():\n";
  
  if (rank < 1)
    cerr << fnName
	 << "  Bad size";

  cartanMat = intmat(rank, rank, 0);
  for (unsigned int i = 0; i < rank; ++i) {
    cartanMat(i, i) = 2;
    if (i > 0)
      cartanMat(i, i-1) = -1;
    if (i < rank-1)
      cartanMat(i, i+1) = -1;
  }
  // simpleRootLength = intvec(rank, 1);

  bool good = true;
  switch (type) {
  case 'A':
    good = true;
    bilinearForm = cartanMat;
    numPositiveRoots = (rank * (rank + 1)) / 2;
    break;
   
  case 'B':
    if (good = (rank >= 2)) {
      cartanMat(rank-1, rank-2) = -2;
      // simpleRootLength(rank-1) = 2;
      bilinearForm = cartanMat;
      // multiply last column by 2
      bilinearForm(rank-2, rank-1) = -2;
      bilinearForm(rank-1, rank-1) = 4;
      numPositiveRoots = rank * rank;
    }
    break;
  
  case 'C':
    if (good = (rank >= 2)) {
      cartanMat(rank-2, rank-1) = -2;
      // simpleRootLength = intvec(rank, 2);
      // simpleRootLength(rank-1) = 1;
      bilinearForm = cartanMat;
      // multiply all cols by 2 then
      // divide last by 2
      bilinearForm *= 2;
      bilinearForm(rank-2, rank-1) = -2;
      bilinearForm(rank-1, rank-1) = 1;
      numPositiveRoots = rank * rank;
    }
    break;
  
  case 'D':
    if (good = (rank >= 4)) {
      cartanMat(rank-3,rank-1) = -1;
      cartanMat(rank-2,rank-1) = 0;
      cartanMat(rank-1,rank-3) = -1;
      cartanMat(rank-1,rank-2) = 0;
      bilinearForm = cartanMat;
      numPositiveRoots = rank * (rank - 1);
    }
    break;
  
  case 'E':
    if (good = (rank >= 6 && rank <= 8)) {
      cartanMat(rank-4,rank-1) = -1;
      cartanMat(rank-2,rank-1) = 0;
      cartanMat(rank-1,rank-4) = -1;
      cartanMat(rank-1,rank-2) = 0;
      bilinearForm = cartanMat;
      if (rank == 6) numPositiveRoots = 35;
      else if (rank == 7) numPositiveRoots = 63;
      else numPositiveRoots = 120;
    }
    break;
  
  case 'F':
    if (good = (rank == 4)) {
      cartanMat(2,1) = -2;
      // simpleRootLength(2) = simpleRootLength(3) = 2;
      bilinearForm = cartanMat;
      // multiply last two col by 2
      bilinearForm(1,2) = -2;
      bilinearForm(2,2) = 4;
      bilinearForm(3,2) = -2;
      bilinearForm(2,3) = -2;
      bilinearForm(3,3) = 4;
      numPositiveRoots = 24;
    }
    break;
  
  case 'G':
    if (good = (rank == 2)) {
      cartanMat(1,0) = -3;
      bilinearForm = cartanMat;
      // multiply last col by 3
      bilinearForm(0,1) = -3;
      bilinearForm(1,1) = 6;
      // simpleRootLength(1) = 3;
      numPositiveRoots = 6;
    }
    break;
  
  default:
    good = false;
  }
  
  if (!good)
    BAD_ARG(fnName);
}

// }}}
// {{{ rootSystem::buildRoots()

void rootSystem::buildRoots() {
  static char * fnName = "rootSystem::buildRoots()";

  positiveRoot = new intvec[numPositiveRoots];

  rootReflection.setLen(numPositiveRoots);
  for (unsigned int i = 0; i < rank; ++i) {
    positiveRoot[i] = simpleRoot[i];
    rootReflection[i] = i + 1;
  }

  unsigned int
    gotRoots = rank,
    newRoots = rank,
    prevGotRoots = 0;

  while (gotRoots < numPositiveRoots) {
    for (unsigned int i = prevGotRoots; i < gotRoots; ++i)
      for (unsigned int j = 1; j <= rank; ++j) {
	// new candidate obtained by multiplying the ith
	// previous candidates by the jth simple reflection
	intvec v(positiveRoot[i] * weylMat[j]);
	bool gotOne = ((rho * cartanMat * v) > 0);
	if (gotOne)
	  for (unsigned int k = 0; k < newRoots; ++k)
	    if (v == positiveRoot[k]) {
	      gotOne = false;
	      break;
	    }
	if (gotOne) {
	  // found a new positive root
#ifdef EXTRA_SAFE
	  if (newRoots >= numPositiveRoots)
	    PANIC(fnName, "Found too many positive roots");
#endif
	  positiveRoot[newRoots] = v;
	  rootReflection[newRoots++]
	    = (j * weylPerm[rootReflection[i]]) * weylPerm[j];
	}
      }
    prevGotRoots = gotRoots;
    gotRoots = newRoots;
  }
  
  // Sanity checks
  if (newRoots < numPositiveRoots)
    PANIC(fnName, "Didn't find enough positive roots.");

  // hunt the highest short root
  int height = 0;
  
  for (unsigned int i = 0; i < numPositiveRoots; ++i) {
    intvec &u = positiveRoot[i];
    // short roots have length 1, but the bilnear form
    // matrix has been scaled by a factor of 2 to make it integral
    if (((u * bilinearForm * u) == 2)
	&& (u.sum() > height)) {
      highestShortRoot = i;
      height = u.sum();
    }
  }

  // sanity check
  if (height == 0)
    PANIC(fnName, "Couldn't find the highest short root.");
}

// }}}
// {{{ rootSystem::buildCoxeterGenerators()

void rootSystem::buildCoxeterGenerators() {
  // s is the set of Coxeter generators of the affine Weyl

  s = new affineWeyl[rank + 1];
    
  // the ordinary simple reflections
  for (unsigned int i = 1; i <= rank; ++i)
    s[i] = affineWeyl(this, i, origin);
  
  // the affine simple reflection
  int refl = rootReflection[highestShortRoot];
  intvec & v = fundamentalVertex[1];
  s[0] = affineWeyl(this, refl, v - (v * weylMat[refl]));
}

// }}}
// {{{ rootSystem::buildWeylMat()

void rootSystem::buildWeylMat() {
  // Generate the matrices of the ordinary weyl group
  // The non-affine simple reflection matrices
  // The list of matrices weylMat is ordered compatibly with length.
  // In particular, the identity matrix is weylMat[0],
  // and the simple reflections are weylMat[1], ... ,weylMat[rank]

  intmat* w[numPositiveRoots + 1];
  // Temporary storage for the construction
  // w[i] will be the array of Weyl elements of length i

  unsigned int numLength[numPositiveRoots + 1];
  // numLength[i] will be the number of Weyl elements of length i

  // Identity mat
  w[0] = new intmat[1];
  w[0][0] = identityMat;
  numLength[0] = 1;
    
  // Generate the simple reflections
  w[1] = new intmat[rank];
  for (unsigned int i = 0; i < rank; ++i) {
    w[1][i] = identityMat;
    for (unsigned int j = 0; j < rank; ++j)
      w[1][i](j,i) -= cartanMat(j,i);
  }
  numLength[1] = rank;

  weylSize = rank + 1;

  // Now inductively build the rest of the Weyl group
  for (unsigned int i = 1; i < numPositiveRoots; ++i) {
    // This is an upper bound on how many we can generate.
    // Usually it's way over the top.
    w[i+1] = new intmat[numLength[i] * rank];
    numLength[i+1] = 0;
    for (unsigned int j = 0; j < numLength[i]; ++j)
      for (unsigned int k = 0; k < rank; ++k) {
	intmat m = w[i][j] * w[1][k];
	// Now check if we have it already.
	// We only have to look at lengths i-1 and i+1
	int newMat = true;
	for (unsigned int l = 0; l < numLength[i-1]; ++l)
	  if (m == w[i-1][l]) {
	    newMat = false;
	    break;
	  }
	if (newMat)
	  for (unsigned int l = 0; l < numLength[i+1]; ++l)
	    if (m == w[i+1][l]) {
	      newMat = false;
	      break;
	    }
	if (newMat)
	  // It's new
	  w[i+1][numLength[i+1]++] = m;
      }
    weylSize += numLength[i+1];
  }

  // Sanity check
  if (numLength[numPositiveRoots] != 1)
    PANIC("rootSystem::buildWeylMat()",
	  "construction of ordinary Weyl group failed.");

  // Generation complete, so decant the information into
  // permanent storage, and release temporary space.
  weylMat = new intmat[weylSize];
  weylLength.setLen(weylSize);
  
  for (unsigned int i = 0, k = 0; i <= numPositiveRoots; ++i) {
    for (unsigned int j = 0; j < numLength[i]; ++j) {
      weylMat[k] = w[i][j];
      weylLength[k++] = i;
    }
    delete [] w[i];
  }
}

// }}}
// {{{ rootSystem::buildWeylPerm()

void rootSystem::buildWeylPerm() {
  weylPerm = new perm[weylSize];
  weylInv.setLen(weylSize);

  intvec v(weylSize); // temporary scratch
  for (unsigned int i = 0; i < weylSize; ++i) {
    intmat &w = weylMat[i];
    for (unsigned int j = 0; j < weylSize; ++j) {
      intmat x(weylMat[j] * w);
      intvec unchecked(weylSize, 1);
      unsigned int k;
      for (k = 0; k < weylSize; ++k) {
	if (unchecked[k] && (x == weylMat[k])) {
	  if (k == 0)
	    weylInv[i] = j;
	  break;
	}
      }
      
#ifdef EXTRA_SAFE
      if (k >= weylSize)
	PANIC("rootSystem::buildWeylPerm()",
	      "couldn't find the product.");
#endif
 
      unchecked[k] = 0;
      v[j] = k;
    }
    perm foo(v);
    weylPerm[i] = foo;
  }
}

// }}}
// {{{ rootSystem::buildAlcoves

void rootSystem::buildAlcoves() {

  // vertices of the fundamental alcove
  intvec &HSR = positiveRoot[highestShortRoot];
  intvec adjust(rank);

  for (unsigned int i = 0; i < rank; ++i)
    adjust[i] = fundamentalWeight[i] * bilinearForm * HSR;
  adjust /= hcf(adjust);
  prime = lcm(adjust);
  
  fundamentalVertex = new intvec[rank + 1];
  fundamentalVertex[0] = origin;
  alcoveCentre = new intvec[weylSize]; // centres of alcoves corresponding to W
  intvec &centre = alcoveCentre[0];
  centre = origin;

  int tmp = prime;
  for (unsigned int i = 1; i <= rank; ++i) {
    fundamentalVertex[i]
      = (fundamentalWeight[i-1] * (prime / adjust[i-1]));
    tmp = hcf(tmp, hcf(fundamentalVertex[i]));
  }

  // Make the alcove as small as we can
  prime /= tmp;
  for (unsigned int i = 1; i <= rank; ++i) {
    fundamentalVertex[i] /= tmp;
    centre += fundamentalVertex[i];
  }

  // Want to divide centre by rank + 1 to get the true centre.
  // This may take some retroactive scaling to keep things integral:
  tmp = hcf(hcf(centre), rank+1);
  centre /= tmp;
  tmp = (rank + 1) / tmp;
  prime *= tmp;
  for (unsigned int i = 0; i < rank; ++i)
    fundamentalVertex[i+1] *= tmp;

  // Compute centres of the alcoves in Weyl orbit of fundamental alcove
  for (unsigned int i = 1; i < weylSize; ++i)
    alcoveCentre[i] = (centre * weylMat[i]);

  // Lists indices of the vertices of the faces
  // of the fundamental alcove.
  // These are used to produce data files readable
  // by Geomview.

  alcoveFace = new intvec[rank + 1];
  for (unsigned int i = 0; i <= rank; ++i) {
    alcoveFace[i].setLen(rank);
    for (unsigned int j = 0, k = 0; j <= rank; ++j)
      if (j != i)
	alcoveFace[i][k++] = j;
  }

  // The canonical centres are the alcove centres
  // with coefficients between 0 and prime.
  // They are used for converting from alcove centres
  // to affine Weyl group elements.
  // Each is a translate by prime * root lattice of
  // the corresponding ordinary Weyl group centre

  canonicalCentre = new intvec[weylSize];
  for (unsigned int i = 0; i < weylSize; ++i) {
    intvec v = alcoveCentre[i];
    for (unsigned int j = 0; j < rank; ++j)
      if (v[j] < 0)
	v[j] += prime;
    canonicalCentre[i] = v;
  }

}

// }}}
// {{{ rootSystem::gvPrintAxes

// The caller has to add preamble, postamble and flush the stream.

void rootSystem::gvPrintAxes(ostream* s, int mag) const {
  unsigned int i, j;
  *s << "VECT "
     << rank << " " << (2 * rank) << " 0\n"; // num lines, vertices, colours
  for (i = rank; i--;)
    *s << "2 "; // num vertices in each line
  *s << "\n";
  for (i = rank; i--;)
    *s << "0 "; // num colours in each line
  *s << "\n";

  intvec& origin = fundamentalVertex[0];

  for (i = 1; i <= rank; ++i) {
    intvec pt(fundamentalVertex[i] * mag);
    for (j = 0; j < rank; ++j)
      *s << origin[j] << " ";
    for (; j < 3; ++j)
      *s << "0 ";
    *s << "\n";
    for (j = 0; j < rank; ++j)
      *s << pt[j] << " ";
    for (; j < 3; ++j)
      *s << "0 ";
    *s << "\n";
  }
}

// }}}

// {{{ UNINCORPORATED STUFF FROM GAP

//    # Make it a set for efficient lookup, but keep the indices so
//    # we can find the base alcoves from which they came.
//  
//    RR.canonicalIndex := [1..RR.WeylSize];
//    SortParallel(RR.canonicalCentre, RR.canonicalIndex);
//  
//    # }}}
//    # {{{ Compute Alcove Orientations
//  
//    # Choose a provisional set of representatives of the set of alcove
//    # orientations.
//    
//    RR.orientation := [];
//    foo := [];
//  
//    for i in [1..RR.WeylSize] do
//  	notfound := true;
//  	j := 1;
//  	while notfound and j <= Length(foo) do
//  	  if AffineWeylAreTranslates([i, RR.origin],
//  		     [foo[j], RR.origin]) then
//  	    notfound := false;
//  	    RR.orientation[i] := foo[j];
//  	  fi;
//  	  j := j + 1;
//  	od;
//  	if notfound then
//  	  Add(foo, i);
//  	  RR.orientation[i] := i;
//  	fi;
//    od;
//    RR.numOrientations := Length(foo);
//    
//    # Now replace this rather arbitrary choice by those orientations
//    # encountered on minimal walks upward from the fundamental alcove.
//    # At the same time we compute the `walk graph' amongst such alcoves.
//    
//    # We build:
//    # RR.WalksUp:    List of ways up from a given orientation
//    # RR.WalksDown:  List of ways down from a given orientation
//    
//    # Start with the fundamental alcove.
//  
//    alcoves := [];
//    alcoves[1] := [[RR.id, RR.origin]];
//    ReplaceOrientation(RR.orientation[RR.id], RR.id);
//    RR.WalksUp := [];
//    RR.WalksDown := [];
//    RR.boxAlcove := [];
//    RR.WalksUp[RR.id] := [];
//    RR.WalksDown[RR.id] := [];
//    RR.boxAlcove[RR.id] := [RR.id, RR.origin];
//    RR.orientationRep := [RR.id]; # list of orientations encountered thus far
//    
//    # Generate the remaining orientations by walking upwards across
//    # walls into the dominant chamber from the fundamental alcove.
//    
//    todo := RR.numOrientations - 1;
//    level := 1;
//    
//    while todo > 0 do
//  	# Run through the alcoves at this level
//  	# generating those at the next as we go.
//  	
//  	alcoves[level+1] := []; # Alcoves found at next level
//  	NewOrientations := [];  # Orientations found at next level      
//  	
//  	for a in alcoves[level] do
//  	  # Run through the types of wall
//  	  for i in [1..Length(RR.s)] do
//  	    
//  	    # If a is below the s-wall, and the alcove above it
//  	    # has a new orientation, collect its data
//  	    
//  	    b := AffineWeylProduct(RR.s[i], a);
//  	    bidx := RR.orientation[b[1]];
//  	    if AffineWeylDominates(b, a) and
//  	       not (bidx in RR.orientationRep) then
//  	      if not bidx in NewOrientations then
//  		ReplaceOrientation(bidx, b[1]);
//  		RR.boxAlcove[b[1]] := b;
//  		Add(NewOrientations, b[1]);
//  		Add(alcoves[level+1], b);
//  		RR.WalksUp[b[1]] := [];
//  		RR.WalksDown[b[1]] := [[i, a[1]]];
//  		Add(RR.WalksUp[a[1]], [i, b[1]]);
//  		todo := todo -1;
//  	      else
//  		Add(RR.WalksDown[bidx], [i, a[1]]);
//  		Add(RR.WalksUp[a[1]], [i, bidx]);
//  	      fi;
//  	    fi;
//  	  od;
//  	od;
//  	
//  	Append(RR.orientationRep, NewOrientations);
//  	level := level + 1;
//    od;
//  
//    # }}}
//    # {{{ Information for recursively computing arbitrary tilting chars
//  
//    RR.KnownWeyl := [[RR.id, RR.origin]];
//    RR.KnownIdx := [1];
//    RR.KnownTilt := [[[RR.id],[RR.origin],[v^0]]];
//  
//    # }}}
//    
//    SaveTo(CM, RR);
//  end;
//  
//  CMSet := BuildRootRecord;

// }}}

// }}}

C_END(rootSystem);

// {{{ Local Variables

// Local Variables:
// mode: c++
// folded-file: t
// End:

// }}}

