WARNING;
// Copyright (C) 2000 Paul Martin and Dave Woodcock. All rights reserved.
// $Id: tiltChar.m4,v 1.7 2000/11/13 06:16:40 dave Exp $

// This file defines a single function, tiltChar, which implements
// Wolfgang Soergel's celebrated algorithm for computing the Weyl
// module multiplicities in quantum tilting modules. This function is
// the raison d'tre for the entire libtilt package.
//
// For further details see W. Soergel, "Kazhdan-Lusztig polynomials
// and a combinatoric for tilting modules", Representation Theory 1
// (1997), 83-114.
//
// Note:
//   #define VERBOSE_TILT_CHAR
// gives diagnostic information when building tilting characters.
// (We probably need to change invocations of cout to cerr to use this now.)
//
#define KL_POLYNOMIALS_TO_FILE
// causes all computed polynomials to be sent to a file, hardwired below to be
// /tmp/tiltChars/sanity  (for whatever reason!). 

H_BEGIN(tiltChar);

#include "alcove.h"
#include "character.h"

const character* tiltChar(Alcove*);

H_END(tiltChar);
C_BEGIN(tiltChar);

#include "exception.h"
#include <fstream.h>
// PPM: This last added for file writing.

// {{{ tiltChar()

const character* tiltChar(Alcove* a) {
  static char* fnName = "alcove::tiltChar(): \n";
  character* result;

  // PPM: outfname will just be the hardwired name (set below) of a file 
  // to write parabolic Kazhdan-Lusztig polynomials to, if required. 
  // If you dont need to know polynomials you can safely comment out this line.
  char outfname[80];

  if ((result = (character*)(a->tiltChar()))) return result;
  if (!a->isDominant())
    BAD_ARG(fnName);
 
  // {{{ VERBOSE_TILT_CHAR: "Calculating"
#ifdef VERBOSE_TILT_CHAR
  cout << "Calculating tilting char for " << *a << "\n";
#endif
  // }}}

  // Scan for a downward alcove, preferably one
  // whose character has already been computed

  Alcove *b = 0;      // The alcove we reflect down to...
  unsigned int l = 0; // ..and the reflection we use to get there

  for (unsigned int j = 0; j <= a->rank(); ++j) {
    if (a->reflectsDown(j)) {
      // Either we take the first downward reflection whose
      // associated character has already been computed,
      // or we take the last downward reflection we find
      l = j;
      b = a->reflection(j);
      if (b->tiltChar()) break;
    }
  }

  if (b) {
    // {{{ VERBOSE_TILT_CHAR: "reflecting"
#ifdef VERBOSE_TILT_CHAR
      cout << "Reflecting in " << l << "\n"
	   << "Old alcove " << *a << "\n"
	   << "New alcove " << *b << "\n";
#endif
      // }}}
    result = tiltChar(b)->reflection(l); // creates a new character
      
    // Cancel summands
    for (unsigned int k = result->len(); k--;) {
      int constantTerm = result->p(k)(0);
      // {{{ EXTRA_SAFE: check constantTerm >= 0
#ifdef EXTRA_SAFE
	if (constantTerm < 0)
           PANIC(fnName, "Soergel's algorithm dies in flames!");
#endif
	// }}}
      if (constantTerm > 0) {
	Alcove* c = result->a(k);
	if (c != a) {
	  // {{{ VERBOSE_TILT_CHAR: "cancelling"
#ifdef VERBOSE_TILT_CHAR
	    cout << "Cancelling the tilting char for "
		 << *c << " " << constantTerm << " times.\n";
#endif
	    // }}}
	  result->cancel(tiltChar(c), constantTerm);
	  // {{{ VERBOSE_TILT_CHAR: "cancelled"
#ifdef VERBOSE_TILT_CHAR
	    cout << "Cancelled\n";
#endif
// }}}
	}
      }
    }
    // Cancel zeroes from what's left
#ifdef VERBOSE_TILT_CHAR
    cout << "Culling zeroes\n";
#endif
    result->cullZeroes();
#ifdef VERBOSE_TILT_CHAR
    cout << "Culled\n";
#endif

  } else {
    // No downward reflection, so a = fundamental alcove.
    result = new character (a, polhash.addr(intpol(0,1)));
  }

  a->setTiltChar(result);
// {{{ ifdef KL_POLYNOMIALS_TO_FILE

#ifdef KL_POLYNOMIALS_TO_FILE
  // PPM: Writing the characters to a file:
  //  ofstream PPMout("/tmp/tiltChars/take1");   // this would overwrite each char
  // PPM: next few lines try to sanely name the output file(s), 
  // but it is ugly at best!!
  // Maybe later we will try to postpone this write-to-file to the 
  // perl layer where tiltChar is called..
  strcpy(outfname, "/tmp/tiltChars/");
  strcat(outfname,"sanity");
  ofstream PPMout;
  //  PPMout.open("/tmp/tiltChars/take2", ios::app);
  PPMout.open(outfname, ios::app);
  result->print(PPMout);
  PPMout << "\n";
  PPMout.close();
#endif

// }}}
  return result;
}

// }}}

C_END(tiltChar);

// {{{ Local Variables

// Local Variables:
// mode: c++
// folded-file: t
// End:

// }}}
