WARNING;
// Copyright (C) 2000 Paul Martin and Dave Woodcock. All rights reserved.
// $Id: urarray.m4,v 1.4 2000/11/13 06:16:48 dave Exp $
H_BEGIN(urarray);

#include <iostream.h>
#ifdef EXTRA_SAFE
#include "exception.h"
#endif // EXTRA_SAFE
#include "urlist.h"

// {{{ class urarray

template<class X>
class urarray {
private:
  unsigned int wwidth;
  unsigned int hheight;
  X *entry;

#ifdef EXTRA_SAFE
  void verifyRange(unsigned int, unsigned int) const;
#endif
  
public:
  urarray(unsigned int w=0, unsigned int h=0);
  urarray(unsigned int w, unsigned int h, const X& val); // sets all entries to val
  urarray(const urarray<X>&);
  ~urarray();

  urarray<X>& operator=(const urarray<X>&);
  void setDim(unsigned int i, unsigned int j);
  X& operator() (unsigned int i, unsigned int j);
  const X& operator() (unsigned int i, unsigned int j) const;

  bool operator== (const urarray<X>&) const;
  bool operator!= (const urarray<X>& a) const {return !operator==(a);}
  
  unsigned int width() const {return wwidth;};
  unsigned int height() const {return hheight;};
  urlist<X> row(unsigned int i) const;
  urlist<X> col(unsigned int i) const;
  ostream& print(ostream&) const;
};

// }}}
// {{{ inline member functions

// {{{ urarray<X>::operator()

template<class X>
inline X& urarray<X>::operator() (unsigned int i, unsigned int j) {
#ifdef EXTRA_SAFE
  verifyRange(i,j);
#endif
  return entry[wwidth * i + j];
}

template<class X>
inline const X& urarray<X>::operator() (unsigned int i, unsigned int j) const {
#ifdef EXTRA_SAFE
  verifyRange(i,j);
#endif
  return entry[wwidth * i + j];
}

// }}}
// {{{ operator<< ()

template<class X>
inline ostream& operator<< (ostream& s, const urarray<X>& m) {
  return m.print(s);
}

// }}}
// {{{ urarray<X>::setDim

template<class X>
inline void urarray<X>::setDim(unsigned int h, unsigned int w) {
  unsigned int r, s;
  if ((r = hheight * wwidth) != (s = h * w)) {
    if (r) delete [] entry;
    if (s) entry = new X [s];
  }
  hheight = h;
  wwidth = w;
}

// }}}

// }}}
// {{{ urarray member functions

// {{{ urarray<X>::verifyRange()

#ifdef EXTRA_SAFE
template<class X>
void urarray<X>::verifyRange(unsigned int i, unsigned int j) const {
  if ((i >= hheight) || (j >= wwidth))
    RANGE_ERROR("urarray::verifyRange()");
}
#endif

// }}}
// {{{ urarray<X>::urarray<X>()

// {{{ (unsigned int, unsigned int)

template<class X>
urarray<X>::urarray(unsigned int h, unsigned int w)
  : wwidth(w), hheight(h)
{
  static unsigned int n;
  if ((n = hheight * wwidth))
    entry = new X [n];
}

// }}}
// {{{ (int, int, const X&)

template<class X>
urarray<X>::urarray(unsigned int h, unsigned int w, const X& val)
  : wwidth(w), hheight(h)
{
  static unsigned int n;
  if ((n = hheight * wwidth)) {
    entry = new X [n];
    for (unsigned int i = 0; i < n; ++i)
      entry[i] = val;
  }
}

// }}}
// {{{ (const urarray<X>&)

template<class X>
urarray<X>::urarray(const urarray<X>& m)
  : wwidth(m.wwidth), hheight(m.hheight)
{
  // copy constructor
  static unsigned int n;
  if ((n = hheight * wwidth)) {
    entry = new X [n];
    for (unsigned int i = 0; i < n; ++i)
      entry[i] = m.entry[i];
  }
}

// }}}

// }}}
// {{{ urarray<X>::~urarray<X>()

template<class X>
urarray<X>::~urarray() {
  if (hheight * wwidth)
    delete [] entry;
}

// }}}
// {{{ urarray<X>::operator=()

template<class X>
urarray<X>& urarray<X>::operator=(const urarray<X>& m) {
  // Harmless if passed itself
  setDim(m.hheight, m.wwidth);
  unsigned int n = wwidth * hheight;
  for (unsigned int i = 0; i < n; ++i)
    entry[i] = m.entry[i];
  return *this;
}

// }}}
// {{{ urarray<X>::operator==()

template<class X>
bool urarray<X>::operator== (const urarray<X>& m) const {

  if ((wwidth != m.wwidth)
      || (hheight != m.hheight))
    return false;

  unsigned int n = wwidth * hheight;
  for (unsigned int i = 0; i < n; ++i)
    if (entry[i] != m.entry[i])
      return false;
  return true;
}

// }}}
// {{{ urarray<X>::row()

template<class X>
urlist<X> urarray<X>::row(unsigned int i) const {

#ifdef EXTRA_SAFE
  if (i >= hheight)
    RANGE_ERROR("urarray::row");
#endif

  urlist<X> v(wwidth);
  for (unsigned int j = 0, k = i * wwidth; j < wwidth;)
    v[j++] = entry[k++];
  return v;
}

// }}}
// {{{ urarray<X>::col()

template<class X>
urlist<X> urarray<X>::col(unsigned int i) const {

#ifdef EXTRA_SAFE
  if (i >= wwidth)
    RANGE_ERROR("urarray::col");
#endif

  urlist<X> v(hheight);
  for (unsigned int j = 0; j < hheight; ++j, i += wwidth)
    v[j] = entry[i];
  return v;
}

// }}}
// {{{ urarray<X>::print()

template<class X>
ostream& urarray<X>::print(ostream& s) const {
  s << "[";
  unsigned int k = 0;
  for (unsigned int i = 0; i < hheight; ++i) {
    if (i > 0)
      s << ",[";
    else
      s << "[";
    for (unsigned int j = 0; j < wwidth; ++j)
      if (j > 0)
	s << "," << entry[k++];
      else
	s << entry[k++];
    s << "]";
  }
  s << "]";
  return s;
}

// }}}

// }}}

H_END(urarray);

// {{{ Local Variables

// Local Variables:
// mode: c++
// folded-file: t
// End:

// }}}
