WARNING;
// Copyright (C) 2000/3 Paul Martin and Dave Woodcock. All rights reserved.
// $Id: urmat.m4,v 1.5 2000/11/13 06:16:50 dave Exp $
H_BEGIN(urmat);

#include <iostream.h>
#include "exception.h"
#include "urvec.h"
#include "urarray.h"
#include "hcf.h"

// {{{ class urmat

template<class X>
class urmat : public urarray<X> {
public:
  // deferring to urarray
  urmat(unsigned int w=0, unsigned int h=0) : urarray<X>(w,h) {};
  urmat(unsigned int w, unsigned int h, const X& x) : urarray<X>(w,h,x) {};
  urmat(const urarray<X>&a) : urarray<X>(a) {};

  urmat<X>& operator=(const urarray<X>&m) {
    return (urmat<X>&)urarray<X>::operator=(m);
  }
  urvec<X> row(unsigned int i) {
    return (urvec<X>)urarray<X>::row(i);
  }
  urvec<X> col(unsigned int i) {
    return (urvec<X>)urarray<X>::col(i);
  }
  
  // new stuff
  urmat<X>& operator*=(const X&);
  urmat<X>& operator/=(const X&);
  urmat<X>& operator*=(const urmat<X>&);
  urmat<X>& operator+=(const urmat<X>&);
  urmat<X>& operator-=(const urmat<X>&);
  urmat<X>& negate();

  urmat<X> operator+(const urmat<X>& m) const {
    urmat<X> res(*this);
    return res += m;
  }
  urmat<X> operator-() const {
    urmat<X> res(*this);
    return res.negate();
  }
  urmat<X> operator-(const urmat<X>& m) const {
    urmat<X> res(*this);
    return res -= m;
  }
  urmat<X> operator*(const urmat<X>& m) const {
    urmat<X> res(*this);
    return res *= m;
  }
  urmat<X> operator*(const X& a) const {
    urmat<X> res(*this);
    return res *= a;
  }
  urmat<X> operator/(const X& a) const {
    urmat<X> res(*this);
    return res /= a;
  }

  urvec<X> operator*(const urvec<X>&) const;
  urvec<X> rightMultiple(const urvec<X>&) const;
  urvec<X> diagonal() const;
  urmat<X> transposed() const;
  
  void makeIdMat(unsigned int rank);
  void twoRowOp(unsigned int, unsigned int, const urmat<X> &);
  void switchRows(unsigned int, unsigned int);
  void multiplyRow(unsigned int, const X&);
  void addRowMultiple(unsigned int, unsigned int, const X&);

  //  static const X one = X(1);
  //  static const X zero = X(0);
  // PPM:
  static const X one;
  static const X zero;
  static urmat<X> idMat(unsigned int rank);
};
//PPM:
template<class X> const X urmat<X>::zero = X(0);
template<class X> const X urmat<X>::one = X(1);

// }}}
// {{{ global inline derived operators

template<class X>
inline ostream& operator<< (ostream& s, const urmat<X>& m) {
  return m.print(s);
}

template<class X>
inline urvec<X> operator* (const urvec<X>& v, const urmat<X>& m) {
  return m.rightMultiple(v);
}

// }}}
// {{{ urmat member functions

// {{{ urmat<X>::operator+=()

template<class X>
urmat<X>& urmat<X>::operator+= (const urmat<X>& m) {

#ifdef EXTRA_SAFE
  if ((width() != m.width())
      || (height() != m.height()))
    ARG_MISMATCH("urmat::operator+=()");
#endif

  for (unsigned int i = height(); i--;)
    for (unsigned int j = width(); j--;)
      (*this)(i,j) += m(i,j);
  return *this;
}


// }}}
// {{{ urmat<X>::operator-=()

template<class X>
urmat<X>& urmat<X>::operator-= (const urmat<X>& m) {

#ifdef EXTRA_SAFE
  if ((width() != m.width())
      || (height() != m.height()))
    ARG_MISMATCH("urmat::operator-=()");
#endif

  for (unsigned int i = height(); i--;)
    for (unsigned int j = width(); j--;)
      (*this)(i,j) -= m(i,j);
  return *this;
}


// }}}
// {{{ urmat<X>::operator*=()

// {{{ (const X&)

template<class X>
urmat<X>& urmat<X>::operator*= (const X& a) {
  for (unsigned int i = 0; i < height(); ++i)
    for (unsigned int j = 0; j < width(); ++j)
      (*this)(i,j) *= a;
  return *this;
}

// }}}
// {{{ (const urmat<X>&)

template<class X>
urmat<X>& urmat<X>::operator*= (const urmat<X>& m) {

#ifdef EXTRA_SAFE
  if (width() != m.height())
    ARG_MISMATCH("urmat::operator*=()");
#endif

  urmat<X> tmp(height(), width());
  
  for (unsigned int i = height(); i--;)
    for (unsigned int j = m.width(); j--;) {
      X s(0);
      for (unsigned int k = width(); k--;)
	s += (*this)(i, k) * m(k, j);
      tmp(i,j) = s;
    }
  return (*this) = tmp;
}

// }}}

// }}}
// {{{ urmat<X>::operator/=()

template<class X>
urmat<X>& urmat<X>::operator/= (const X& a) {
  for (unsigned int i = 0; i < height(); ++i)
    for (unsigned int j = 0; j < width(); ++j)
      (*this)(i,j) /= a;
  return *this;
}

// }}}
// {{{ urmat<X>::operator* ()

template<class X>
urvec<X> urmat<X>::operator* (const urvec<X>& v) const {

#ifdef EXTRA_SAFE
  if (width() != v.len())
    ARG_MISMATCH("urmat::operator*()");
#endif

  urvec<X> result(height());
  for (unsigned int i = height(); i--;) {
    X s(0);
    for (unsigned int j = width(); j--;)
      s += (*this)(i,j) * v[j];
    result[i] = s;
    }
  return result;
}

// }}}
// {{{ urmat<X>::rightMultiple()

template<class X>
urvec<X> urmat<X>::rightMultiple (const urvec<X>& v) const {

#ifdef EXTRA_SAFE
  if (height() != v.len())
    ARG_MISMATCH("urmat::rightMultiple()");
#endif

  urvec<X> result(width());
  for (unsigned int i = width(); i--;) {
    X s(0);
    for (unsigned int j = height(); j--;)
      s += (*this)(j, i) * v[j];
    result[i] = s;
    }
  return result;
}

// }}}
// {{{ urmat<X>::negate()

template<class X>
urmat<X>& urmat<X>::negate() {
  for (unsigned int i = 0; i < height(); ++i)
    for (unsigned int j = 0; j < width(); ++j)
      (*this)(i, j) = - (*this)(i, j);
  return *this;
}

// }}}
// {{{ urmat<X>::idMat()

template<class X>
urmat<X> urmat<X>::idMat(unsigned int rank) {
  urmat<X> result(rank, rank, zero);
  for (unsigned int i = 0; i < rank; ++i)
    result(i, i) = one;
  return result;
}

// }}}
// {{{ urmat<X>::makeIdMat()

template<class X>
void urmat<X>::makeIdMat(unsigned int rank) {
  setDim(rank, rank);
  for (unsigned int i = 0; i < height(); ++i)
    for (unsigned int j = 0; j < width(); ++j)
      (*this)(i,j) = ((i == j) ? 1 : 0);
}

// }}}
// {{{ urmat<X>::diagonal()

template<class X>
urvec<X> urmat<X>::diagonal() const {
  unsigned int min = width();
  if (min > height())
    min = height();
  urvec<X> v(min);
  for (unsigned int i = 0; i < min; ++i)
    v[i] = (*this)(i,i);
  return v;
}

// }}}
// {{{ urmat<X>::switchRows()

template<class X>
void urmat<X>::switchRows(unsigned int a, unsigned int b) {
  
#ifdef EXTRA_SAFE
  if ((a >= height())
      || (b >= height())
      || (a == b))
    RANGE_ERROR("urmat::switchRows()");
#endif

  for (unsigned int i = width(); i--;) {
    X tmp((*this)(b,i));
    (*this)(b,i) = (*this)(a, i);
    (*this)(a,i) = tmp;
  }
}

// }}}
// {{{ urmat<X>::multiplyRow()

template<class X>
void urmat<X>::multiplyRow(unsigned int a, const X& x) {

#ifdef EXTRA_SAFE
  if (a >= height())
    RANGE_ERROR("urmat::multiplyRow()");
#endif

  for (unsigned int i = width(); i--;)
    (*this)(a,i) *= x;
}

// }}}
// {{{ urmat<X>::addRowMultiple()

template<class X>
void urmat<X>::addRowMultiple(unsigned int a, unsigned int b, const X& x) {
  
#ifdef EXTRA_SAFE
  if ((a >= height())
      || (b >= height())
      || (a == b))
    RANGE_ERROR("urmat::addRowMultiple()");
#endif

  for (unsigned int i = width(); i--;)
    (*this)(a, i) += (*this)(b, i) * x;
}

// }}}
// {{{ urmat<X>::twoRowOp()

template<class X>
void urmat<X>::twoRowOp(unsigned int a, unsigned int b, const urmat<X>& m) {
  // Multiply the submatrix on the rows [a,b] by m
  // This is a generalised row operation, used to compute
  // inverses.

#ifdef EXTRA_SAFE
  static char* fnName = "urmat::twoRowOp()";
  if ((m.height() != 2) || m.width() != 2)
    BAD_ARG(fnName);
  
  if ((a >= height())
      || (b >= height())
      || (a == b))
    RANGE_ERROR(fnName);
#endif

  for (unsigned int i = 0; i < width(); ++i) {
    X tmp (m(0,0) * (*this)(a,i) + m(0,1) * (*this)(b,i));
    (*this)(b,i) = m(1,0) * (*this)(a,i) + m(1,1) * (*this)(b,i);
    (*this)(a,i) = tmp;
  }
}

// }}}
// {{{ urmat<X>::transposed()

template<class X>
urmat<X> urmat<X>::transposed() const {
  urmat<X> res(width(), height());
  for (unsigned int i = 0; i < height(); ++i)
    for (unsigned int j = 0; j < width(); ++j)
      res(j,i) = (*this)(i,j);
  return res;
}

// }}}

// }}}

// {{{ GramSchmidt()

template<class X>
void GramSchmidt(urmat<X>& m, const urmat<X>& form) {
  // orthonormalise the rows of the matrix m with respect
  // to the bilinear form given by matrix form
  // assumes existence of a function
  //              X sqrt(const X&)

  unsigned int n = m.width();

  if ((n != m.height())
      || (n != form.width())
      || (n != form.height()))
    ARG_MISMATCH("GramSchmidt()");
  
  for (unsigned int i = 0; i < n; ++i) {
    // subtract suitable multiples of earlier rows
    for (unsigned int j = 0; j < i; ++j) {
      X s(m.row(i) * form * m.row(j));
      for (unsigned int col = 0; col < n; ++col)
	m(i, col) -= s * m(j, col);
    }

    // normalise
    X s(sqrt(m.row(i) * form * m.row(i)));
    for (unsigned int col = 0; col < n; ++col)
      m(i, col) /= s;
  }

  // Sanity check
  // cout << "Gram Schmidt sanity check:\n"
  //      << "Orthoganlised stuf: " << m * form * m.transposed() << "\n";
}

// }}}
// {{{ inv()

template<class X>
urmat<X> inv(const urmat<X>& m) {
  static char* fnName = "inv()";
  unsigned int rank;
  X scaling;
  
  if ((rank = m.height()) != m.width())
    BAD_ARG(fnName);

  urmat<X> tmp(m); // a copy so that we don't corrupt m 
  urmat<X> res(rank, rank);
  res.makeIdMat(rank);
  
  for (unsigned int col = 0; col < rank; ++col) {
    unsigned int row;
    for (row = col; row < rank; ++row)
      if (tmp(row, col) != urmat<X>::zero)
	break;

    if (row == rank)
      BAD_ARG(fnName);

    scaling = X(1) / tmp(row, col);
    tmp.multiplyRow(row, scaling);
    res.multiplyRow(row, scaling);

    if (row > col) {
      tmp.switchRows(col, row);
      res.switchRows(col, row);
      row = col;
    }

    // adjust other rows
    for (unsigned int i = 0; i < rank; ++i)
      if ((i != row) && (tmp(i, col) != 0)) {
	scaling = - tmp(i,col) / tmp(row,col);
	tmp.addRowMultiple(i, row, scaling);
	res.addRowMultiple(i, row, scaling);
      }
  }
  return res;
}

// }}}
// {{{ hcfMat()

template<class X>
urmat<X> hcfMat(const X& u, const X& v) {
  // returns unimodular mx m s.t.
  // m * [u,v]^T = [hcf(u,v), 0]^T

  urmat<X> result(2, 2);
  result.makeIdMat(2);
  static urmat<X> change(2, 2, urmat<X>::one);
  change(0,0) = urmat<X>::zero;
  X uu(u), vv(v);

  while (vv != urmat<X>::zero) {
    change(1,1) = - uu / vv;
    X a = vv;
    vv = uu % vv;
    uu = a;
    result = change * result;
  }
  return result;
}

// }}}
// {{{ quasiInv()

template<class X>
urmat<X> quasiInv(const urmat<X>& m) {

  // Return m' so that m * m' is a (non-zero) scalar matrix.
  // We assume that <X> is a Euclidean domain,
  // in as much as that we assume the following functions to be defined:
  //
  //       urmat<X> hcfMat(const X&, const X&)
  //       X hcf(const X&, const X&)
  //       X lcm(const urvec<X>&)

  static char* fnName = "quasiInv():\n";
  unsigned int rank;
  
  if ((rank = m.height()) != m.width())
    BAD_ARG(fnName);

  urmat<X> tmp(m);
  urmat<X> res(rank, rank);
  res.makeIdMat(rank);
  
  for (unsigned int col = 0; col < rank; ++col) {
    unsigned int row;
    for (row = col; row < rank; ++row)
      if (tmp(row,col) != 0)
	break;

    if (row == rank)
      BAD_ARG(fnName);
    
    if (row > col) {
      tmp.switchRows(col, row);
      res.switchRows(col, row);
      row = col;
    }

    // adjust subsequent rows
    for (unsigned int i = row+1; i < rank; ++i)
      if (tmp(i,col) != urmat<X>::zero) {
	urmat<X> x(hcfMat(tmp(row,col), tmp(i,col)));
	tmp.twoRowOp(row, i, x);
	res.twoRowOp(row, i, x);
      }

    // adjust previous rows
    for (unsigned int i = 0; i < row; ++i)
      if (tmp(i,col) != urmat<X>::zero) {
	X h(hcf(tmp(i,col), tmp(row,col)));
	urmat<X> x(2,2);
	x(0,0) = urmat<X>::one;
	x(0,1) = urmat<X>::zero;
	x(1,0) = - tmp(i,col) / h;
	x(1,1) = tmp(row,col) / h;
	tmp.twoRowOp(row, i, x);
	res.twoRowOp(row, i, x);
      }
  }

  X l(lcm(tmp.diagonal()));
  for (unsigned int i = 0; i < rank; ++i)
    for (unsigned int j = 0; j < rank; ++j)
      res(i,j) *= (l / tmp(i,i));
  return res;
};

// }}}

H_END(urmat);

// {{{ Local Variables

// Local Variables:
// mode: c++
// folded-file: t
// End:

// }}}
