WARNING;
// Copyright (C) 2000 Paul Martin and Dave Woodcock. All rights reserved.
// $Id: urvec.m4,v 1.4 2000/11/20 23:51:27 dave Exp $
H_BEGIN(urvec);

// {{{ doc

// vector template
// Assumes that the parameter class X is a ring,
// i.e. supports the following functions:

// X::X(int);
// X& X::X(const X&);
// X& X::operator+=(const X&);
// X& X::operator-=(const X&);
// X& X::operator*=(const X&);
// X& X::operator/=(const X&);
// X X::operator+(const X&);
// X X::operator-(const X&);
// bool X::operator==(const X&);
// int comp(const X& u, const X& v); // pos if u > v, zero if u == v, neg if u < v

// }}}

#include <iostream.h>
#ifdef EXTRA_SAFE
#include "exception.h"
#endif
#include "urlist.h"

// {{{ class urvec

template<class X>
class urvec : public urlist<X> {
private:
#ifdef EXTRA_SAFE
  void checkSameLen(const urvec<X>&) const;
#endif
  
public:			
  // deferring to urlist<X>
  urvec(int i=0) : urlist<X>(i) {};  	        // allocates space; doesn't clear it
  urvec(int i, const X& x) : urlist<X>(i,x) {}; // initialises i entries to j
  urvec(const urlist<X> &v) : urlist<X>(v) {};
  urvec<X>& operator= (const urlist<X>&v) {
    return (urvec<X>&)urlist<X>::operator=(v);
  }

  // new stuff
  X operator* (const urvec<X>&) const;          // standard inner product
  X sum() const;                     	        // sum of the entries

  // Arithmetic modifiers
  urvec<X>& operator+= (const urvec<X>&);
  urvec<X>& operator-= (const urvec<X>&);
  urvec<X>& operator*= (const X&);
  urvec<X>& operator/= (const X&);
  urvec<X>& negate();
  void makeZero(int i);                         // len = i, entries = 0

  urvec<X> operator+ (const urvec<X>& v) const;
  urvec<X> operator- (const urvec<X>& v) const {
    urvec<X> res(*this);
    return res -= v;
  }
  urvec<X> operator- () const {
    urvec<X> res(*this);
    return res.negate();
  }
  urvec<X> operator* (int a) const {
    urvec<X> res(*this);
    return res *= a;
  }
  urvec<X> operator/ (int a) const {
    urvec<X> res(*this);
    return res /= a;
  }
					   
  // Comparisons			   
  bool operator> (const urvec<X>&) const;       // lex order
  bool operator>= (const urvec<X>&) const;      // lex order
  bool operator< (const urvec<X>& v) const {return !(*this >= v);}
  bool operator<= (const urvec<X>& v) const {return !(*this > v);}

};

template<class X>
inline  urvec<X> urvec<X>::operator+ (const urvec<X>& v) const {
  urvec<X> res(*this);
  return res += v;
}


// }}}
// {{{ inline globals

template<class X>
inline ostream& operator<< (ostream& s, const urvec<X>& v) {
  return v.print(s);
}

// }}}
// {{{ urvec member functions

// {{{ urvec<X>::checkSameLen()

#ifdef EXTRA_SAFE
template<class X>
void urvec<X>::checkSameLen(const urvec<X>&v) const {
  if (len() != v.len())
    ARG_MISMATCH("urvec::checkSameLen()");
}
#endif

// }}}
// {{{ urvec<X>::operator*()

template<class X>
X urvec<X>::operator* (const urvec<X>&v) const {
#ifdef EXTRA_SAFE
  checkSameLen(v);
#endif

  X res(0);
  for (unsigned int i = len(); i--;)
    res += (*this)[i] * v[i];
  return res;
}

// }}}
// {{{ urvec<X>::negate()

template<class X>
urvec<X>& urvec<X>::negate() {
  for (unsigned int i = len(); i--;)
    (*this)[i] = - (*this)[i];
  return *this;
}

// }}}
// {{{ urvec<X>::operator+= ()

template<class X>
urvec<X>& urvec<X>::operator+= (const urvec<X>& v) {
#ifdef EXTRA_SAFE
  checkSameLen(v);
#endif
  for (unsigned int i = len(); i--;)
    (*this)[i] += v[i];
  return *this;
}

// }}}
// {{{ urvec<X>::operator-= ()

template<class X>
urvec<X>& urvec<X>::operator-= (const urvec<X>& v) {
#ifdef EXTRA_SAFE
  checkSameLen(v);
#endif
  for (unsigned int i = len(); i--;)
    (*this)[i] -= v[i];
  return *this;
}

// }}}
// {{{ urvec<X>::operator*=()

template<class X>
urvec<X>& urvec<X>::operator*=(const X& a) {
  for (unsigned int i = len(); i--;)
    (*this)[i] *= a;
  return *this;
}

// }}}
// {{{ urvec<X>::operator/=()

template<class X>
urvec<X>& urvec<X>::operator/= (const X& a) {
  for (unsigned int i = len(); i--;)
    (*this)[i] /= a;
  return *this;
}

// }}}
// {{{ urvec<X>::operator>()

template<class X>
bool urvec<X>::operator> (const urvec<X>& v) const {
#ifdef EXTRA_SAFE
  checkSameLen(v);
#endif
  static int c;
  for(unsigned int i = 0; i < len(); ++i) {
    if ((c = comp((*this)[i], v[i])) > 0) return true;
    else if (c < 0) return false;
  }
  return false;
}

// }}}
// {{{ urvec<X>::operator>=()

template<class X>
bool urvec<X>::operator>= (const urvec<X>& v) const {
#ifdef EXTRA_SAFE
  checkSameLen(v);
#endif
  static int c;
  for(unsigned int i = 0; i < len(); ++i) {
    if ((c = comp((*this)[i], v[i])) > 0) return true;
    else if (c < 0) return false;
  }
  return true;
}

// }}}
// {{{ urvec<X>::makeZero()

template<class X>
void urvec<X>::makeZero(int i) {
  setLen(i);
  if (i==0) return;
  for (int j = 0; j < i; ++j)
    (*this)[j] = zero;
}

// }}}
// {{{ urvec<X>::sum()

template<class X>
X urvec<X>::sum() const {
  X res(0);
  for (unsigned int i = len(); i--;)
    res += (*this)[i];
  return res;
}

// }}}

// }}}

H_END(urvec);
C_BEGIN(urvec);


C_END(urvec);

// {{{ Local Variables

// Local Variables:
// mode: c++
// folded-file: t
// End:

// }}}
