WARNING;
// Copyright (C) 2000 Paul Martin and Dave Woodcock. All rights reserved.
// $Id: intpol.m4,v 1.13 2000/11/13 06:16:53 dave Exp $
// Integer Laurent polynomials in v.
H_BEGIN(intpol);

#include <iostream.h>
#include "exception.h"
#include "intlist.h"
#include "misc.h"

class intpol;
typedef const intpol Intpol;

// {{{ class intpol

class intpol {
  unsigned int num;    			    // number of non-zero monomials
  int* pow;   	       			    /* array of integers of size 2 * num
					     * first n entries are the powers,
					     * next n entries are the coefficients */
  static intlist pBuf;                      // scratch space for powers
  static intlist cBuf;                      // scratch space for coeffs
  intpol(unsigned int i);                   // init with space of length i
  unsigned int multiple(Intpol&) const;
  unsigned int sum(Intpol&, int) const;
  Intpol* nnext;

public:
  intpol() : num(0), pow(0), nnext(0) {};   // init to 0
  intpol(int i, int a);                     // init to a*v^i
  intpol(Intpol&);
  ~intpol() {if (num) delete [] pow;}

  intpol& addMultiple(Intpol &p, int a); // add a*p
  intpol& operator+= (Intpol& p) {return addMultiple(p, 1);}
  intpol& operator-= (Intpol& p) {return addMultiple(p, -1);}
  intpol& operator*= (Intpol&);
  intpol operator+ (Intpol&) const;
  intpol operator- (Intpol&) const;  
  intpol operator* (Intpol&) const;
  intpol operator^ (int) const;
  intpol& operator<< (int i);               // 1 -> v^-i
  intpol& operator>> (int i);               // 1 -> v^i
  intpol operator[] (int i) const;
  int operator() (int i) const;             // coeff of v^i. NB not an lvalue
  bool operator== (Intpol& a) const {
    if (num != a.num) return false;
    for (unsigned int n = num << 1; n--;)
      if (pow[n] != a.pow[n])
	return false;
    return true;
  }
  bool operator!= (Intpol&p) const {return !(operator==(p));}
  intpol& operator= (Intpol&);
  intpol& bar();                            // this is a modifier!

  unsigned long hash() const;
  Intpol* next() const {return nnext;}
  Intpol* setNext(Intpol *p) const {return ((intpol*)(this))->nnext = p;}
  bool isZero() const {return num == 0;}
  ostream& print(ostream&) const;
};

// }}}
// {{{ inline

// {{{ intpol::operator=()

inline intpol& intpol::operator=(Intpol& p) {
  // Harmless if passed itself

  if (num != p.num) {
    if (num) delete [] pow;
    if ((num = p.num)) pow = new int [num << 1];
  }

  register unsigned int n = num << 1;
  while (n--) pow[n] = p.pow[n];
  return *this;
}

// }}}
// {{{ intpol::hash()

inline unsigned long intpol::hash() const {
  unsigned long seed = 1003;
  unsigned int n = num << 1;
  while (n--)
    seed = dirtyHash(seed ^ pow[n]);
  return seed;
}

inline unsigned long hash(const intpol& p) {
  return p.hash();
}

// }}}

// }}}
// {{{ inline derived functions

// {{{ operator<<()

inline ostream& operator<< (ostream& s, Intpol& p) {
  return p.print(s);
}

// }}}
// {{{ operator[]()

inline intpol intpol::operator[] (int i) const {
  intpol res(*this);
  return res >> i;
}

// }}}

// }}}

extern Intpol v;

H_END(intpol);
C_BEGIN(intpol);

#include <math.h>

// {{{ static member functions

intlist intpol::pBuf(64);
intlist intpol::cBuf(64);

// }}}
// {{{ nonstatic member functions

// {{{ intpol::intpol()

intpol::intpol(unsigned int n) : num(n), nnext(0) {
  if (num) pow = new int [num << 1];
}

intpol::intpol(int p, int c) : num(1), nnext(0) {
  pow = new int [2];
  pow[0] = p;
  pow[1] = c;
}

intpol::intpol(Intpol& p) : num(p.num), nnext(0) {
  if (num) {
    unsigned int n = num << 1;
    pow = new int [n];
    while (n--) pow[n] = p.pow[n];
  }
}

// }}}
// {{{ intpol::multiple()

unsigned int intpol::multiple(Intpol& p) const {
  // Internal calculation for products.
  // Return value is the number of non-zero monomials.
  // The coeffs are in cBuf, while pBuf[0] is the offset on the powers.

  if ((num == 0) || p.num == 0) return 0;

  unsigned int
    min = pow[0] + p.pow[0],
    max = pow[num-1] + p.pow[p.num-1],
    range = max - min + 1;
  int
    *coeff = pow + num,
    *pcoeff = p.pow + p.num;
  
  cBuf.clearLen(range);
  
  unsigned int i = num;
  while (i--) {
    unsigned int j = p.num;
    while (j--)
      cBuf[pow[i] + p.pow[j] - min] += (coeff[i] * pcoeff[i]);
  }
  pBuf[0] = min;
  return cBuf.nonzeroCount(range);
}

// }}}
// {{{ intpol::operator*()

intpol intpol::operator*(Intpol& p) const {
  unsigned int n = multiple(p);
  intpol res(n);
  int * coeff = res.pow + n;
  for (unsigned int i = 0, j = 0; j < n; ++i)
    if (cBuf[i]) {
      res.pow[j] = i + pBuf[0];
      coeff[j++] = cBuf[i];
    }
  return res;
}

// }}}
// {{{ intpol::operator*=()

intpol& intpol::operator*= (Intpol& p) {
  unsigned int n = multiple(p);
  if (num != n) {
    if (num) delete [] pow;
    if ((num = n)) pow = new int [num << 1];
  }
  int * coeff = pow + num;
  for (unsigned int i = 0, j = 0; j < num; ++i)
    if (cBuf[i]) {
      pow[j] = i + pBuf[0];
      coeff[j++] = cBuf[i];
    }
  return *this;
}

// }}}
// {{{ intpol::sum()

unsigned int intpol::sum(Intpol& a, int multiple) const {
  cBuf.checkLen(num + a.num);
  pBuf.checkLen(num + a.num);

  unsigned int i = 0, j = 0, k = 0;
  int *coeff = pow + num;
  int *acoeff = a.pow + a.num;

  while (i < num && j < a.num)
    if (pow[i] < a.pow[j]) {
      pBuf[k] = pow[i];
      cBuf[k++] = coeff[i++];
    } else if (pow[i] > a.pow[j]) {
      pBuf[k] = a.pow[j];
      cBuf[k++] = multiple * acoeff[j++];
    } else {
      pBuf[k] = pow[i];
      // don't want to include zero coeffs
      if ((cBuf[k] = coeff[i++] + multiple * acoeff[j++]))
	++k;
    }

  // At most one of the next two while loops gets executed
  while (i < num) {
    pBuf[k] = pow[i];
    cBuf[k++] = coeff[i++];
  }

  while (j < a.num) {
    pBuf[k] = a.pow[j];
    cBuf[k++] = multiple * acoeff[j++];
  }

  return k;
}

// }}}
// {{{ intpol::operator+()

intpol intpol::operator+(Intpol& p) const {
  unsigned int n = sum(p, 1);
  intpol res(n);
  int * coeff = res.pow + n;
  while (n--) {
    res.pow[n] = pBuf[n];
    coeff[n]   = cBuf[n];
  }
  return res;
}

// }}}
// {{{ intpol::operator-()

intpol intpol::operator-(Intpol& p) const {
  unsigned int n = sum(p, -1);
  intpol res(n);
  int * coeff = res.pow + n;
  while (n--) {
    res.pow[n] = pBuf[n];
    coeff[n]   = cBuf[n];
  }
  return res;
}

// }}}
// {{{ intpol::addMultiple()

intpol& intpol::addMultiple(Intpol& a, int multiple) {
  register unsigned int n = sum(a, multiple);
  if (num != n) {
    if (num) delete [] pow;
    if ((num = n)) pow = new int [num << 1];
  }

  int * coeff = pow + num;
  while (n--) {
    pow[n]   = pBuf[n];
    coeff[n] = cBuf[n];
  }
  return *this;
}

// }}}
// {{{ intpol::operator^()

intpol intpol::operator^(int i) const {
  static char* fnName = "intpol::operator^():\n";
  
  if (i > 0) {
    // positive power

    if (num == 1) {
      // Common special case that's easy to do
      // NB pow[1] is the first coefficient
      return intpol(pow[0] * i, (int) ::pow(pow[1], i));

    } else {
      // This is inefficient for large powers.
      // Could easily be rewritten to be much faster.
      intpol result(*this);
      for (int j = 1; j < i; ++j)
	result = result * (*this);
      return result;
    }

  } else if (i == 0) {
    // zero power

    if (num)
      return intpol(0, 1);
    else
      BAD_ARG(fnName);

  } else {
    // negative power
    // only meaningful for monomials with coefficient +-1
    // NB if num == 1, pow[1] is the first coefficient

    int c = 0;
    if ((num != 1)
	|| (((c = pow[1]) != 1) && (c != -1)))
      BAD_ARG(fnName);

    return intpol(pow[0] * i,
		  ((c == 1) || ((i % 2) == 0) ? 1 : -1));
  }
  // We never get here, but the compiler doesn't know that:
  return intpol(0,1);
}

// }}}
// {{{ intpol::operator<<()

intpol& intpol::operator<< (int n) {
  unsigned int i = num;
  while (i--) pow[i] -= n;
  return *this;
}

// }}}
// {{{ intpol::operator>>()

intpol& intpol::operator>> (int n) {
  unsigned int i = num;
  while (i--) pow[i] += n;
  return *this;
}

// }}}
// {{{ intpol::operator() ()

int intpol::operator() (int i) const {
  for (unsigned int j = 0; j < num; ++j)
    if (pow[j] > i)
      return 0;
    else if (pow[j] == i)
      return pow[num + j]; // jth coefficient
  return 0;
}

// }}}
// {{{ intpol::bar()

intpol& intpol::bar() {
  int* coeff = pow + num;
  for (unsigned int i = 0; i < (num / 2); ++i) {
    int tmp = pow[i];
    pow[i] = - pow[num - 1 - i];
    pow[num - 1 - i] = - tmp;
    tmp = coeff[i];
    coeff[i] = coeff[num - 1 - i];
    coeff[num - 1 - i] = tmp;
  }
  return *this;
}

// }}}
// {{{ intpol::print()

ostream& intpol::print(ostream& s) const {
  if (num) {
    for (unsigned int i = 0; i < num; ++i) {
      int p = pow[i];
      int c = pow[num + i]; // ith coefficient

      // output sign
      if (i > 0) {
	if (c > 0) {
	  s << " + ";
	} else {
	  s << " - ";
	  c = -c;
	}
      } else {
	if (c < 0) {
	  s << '-';
	  c = -c;
	}
      }
      
      // output coeff
      if ((c != 1) || (p == 0))
	s << c;

      // output power of v
      if (p == 1)
	s << 'v';
      else if (p != 0)
	s << "v^" << p;
    }
  } else {
    s << '0';
  }
  return s;
}

// }}}

// }}}

Intpol v(1,1);

#ifdef TEST
#undef TEST
// {{{ main()

// For testing only

int main() {
  intpol p1((v^0) + v + (v^2));
  intpol p2((v^-1) - v + (v^3));

  cout << "Testing intpol...\n"

       << "p1 = " << p1 << "\n"
       << "p2 = " << p2 << "\n"
       << "p1[3] = " << p1[3] << "\n"
       << "p1[-2] = " << p1[-2] << "\n"
       << "p1 + p2 = " << (p1 + p2) << "\n"
       << "p1 * p2 = " << (p1 * p2) << "\n";
  
  cout << "First few powers of p1:\n";
  p2 = p1;
  for (int i = 1; i <= 10; ++i) {
    cout << i << ": " << p2 << "\n";
    p2 = (p1 * p2);
  }
}

// }}}
#endif // TEST

C_END(intpol);

// {{{ Local Variables

// Local Variables:
// mode: c++
// folded-file: t
// End:

// }}}
