WARNING;
// Copyright (C) 2000 Paul Martin and Dave Woodcock. All rights reserved.
// $Id: perm.m4,v 1.8 2000/11/13 06:16:55 dave Exp $
// perms act on the right (by *)
H_BEGIN(perm);

#include <iostream.h>
#include "intlist.h"

// {{{ class perm

class perm : private intlist {
private:
  perm(unsigned int i) : intlist(i) {}; // declare space for i entries
  static intlist buf;          // scratch
  static unsigned int crunchedLen(const intlist& v);
  
public:
  perm() : intlist() {};
  perm(const perm& p) : intlist(p) {};
  perm(const intlist&);
  ~perm() {};

  unsigned int rightMultiple(unsigned int) const;
  perm operator* (const perm&) const;
  perm operator^ (int) const;
  bool operator== (const perm &p) const {return intlist::operator==(p);}
  perm& operator=(const perm &p) {return (perm&)(intlist::operator=(p));}
  ostream& print(ostream&) const;

  static const perm id;    // identity perm
  static perm s(unsigned int);      // elementary transpositions
};

// }}}
// {{{ inline functions

// {{{ operator<<()

inline ostream& operator<< (ostream& s, const perm& p) {
  return p.print(s);
}

// }}}
// {{{ perm::rightMultiple()

inline unsigned int perm::rightMultiple(unsigned int a) const {
  if (a >= len()) return a;
  else return (*this)[a];
}

// }}}
// {{{ operator*

inline unsigned int operator*(unsigned int i, const perm& p) {
  return p.rightMultiple(i);
}

// }}}
// {{{ perm::crunchedLen()

inline unsigned int perm::crunchedLen(const intlist& v) {
  unsigned int num = v.len();
  while ((num) && (v[num-1] == int(num-1))) // Strange cast to calm the compiler
    --num;
  return num;
}

// }}}

// }}}

H_END(perm);
C_BEGIN(perm);

#ifdef EXTRA_SAFE
#  include "exception.h"
#endif

// {{{ static members

intlist perm::buf(10);
const perm perm::id(0);

// {{{ perm::s()

perm perm::s(unsigned int i) {
  perm res(i + 2);
  for (unsigned int j = 0; j < i; ++j)
    res[j] = j;
  res[i] = i+1;
  res[i+1] = i;
  return res;
}

// }}}

// }}}
// {{{ perm member functions

// {{{ perm::perm()

perm::perm(const intlist& v) : intlist(v, 0, crunchedLen(v)) {
#ifdef EXTRA_SAFE
  static char* fnName = "perm::perm(intlist&)";
  static unsigned int i, j;
  buf.clearLen(len());

  for (i = 0; i < len(); ++i) {
    j = (*this)[i];
    if (j >= len())
      BAD_ARG(fnName);
    buf[j] = 1;
  }

  for (i = 0; i < len(); ++i)
    if (buf[i] == 0)
      BAD_ARG(fnName);
#endif
}

// }}}
// {{{ perm::operator* ()

perm perm::operator* (const perm& p) const {
  static unsigned int i, j, bound;

  if (len() == p.len()) { // Cancellation can occur
    bound = len() - 1;
    while ((bound > 0) && (p[(*this)[bound]] == (int)bound)) // cast for the compiler
      --bound;
    perm result(bound + 1);
    for (i = 0; i <= bound; ++i)
      result[i] = p[(*this)[i]];
    return result;

  } else if (len() > p.len()) { // No cancellation can occur
    perm result(len());
    for (i = 0; i < len(); ++i)
      if ((j = (*this)[i]) < p.len())
	result[i] = p[j];
      else
	result[i] = j;
    return result;

  } else { // No cancellation can occur
    perm result(p.len());
    for (i = 0; i < len(); ++i)
      result[i] = p[(*this)[i]];
    for (; i < p.len(); ++i)
      result[i] = p[i];
    return result;
  }
}

// }}}
// {{{ perm::operator^()

perm perm::operator^ (int i) const {
  if (i == 0)
    return perm(id);

  else if (i > 0) {
    // !!! This is inefficient !!!
    perm result(*this);
    for (int j = 1; j < i; ++j)
      result = result * (*this);
    return result;

  } else {
    perm res(len());
    for (int j = 0; j < int(len()); ++j) // cast for the compiler
      res[(*this)[j]] = j;
    if ((i = -i) > 1) {
      // !!! This is inefficient !!!
      perm tmp(res);
      for (int j = 1; j < i; ++j)
	res = res * tmp;
    }
    return res;
  }
}

// }}}
// {{{ perm::print()

ostream& perm::print(ostream& s) const {
  if (len()) {
    unsigned int i, j;
    buf.clearLen(len() + 1);  // important to clear one extra!
    while (true) {
      for (i = 0; buf[i]; ++i);     // find next index not yet considered
      if (i >= len()) break;  // we're done
      buf[i] = 1;
      if ((j = (*this)[i]) != i) {
	// start of nontrivial cycle
	s << "(" << i;
	while (j != i) {
	  buf[j] = 1;
	  s << "," << j;
	  j = (*this)[j];
	}
	s << ")";
      }
    }
  } else {
    s << "()";
  }
  return s;
}

// }}}

// }}}

#ifdef TEST
#undef TEST

// {{{ main()

main() {
  perm s0(perm::s(0));
  perm s1(perm::s(1));
  perm s2(perm::s(2));
  perm s3(perm::s(3));
  
  cout << "Testing perm.cc\n"
       << "s0 = " << s0 << "\n"
       << "s1 = " << s1 << "\n"
       << "s2 = " << s2 << "\n"
       << "s3 = " << s3 << "\n"
       << "1 * s0 = " << (1 * s0) << "\n"
       << "5 * s0 = " << (5 * s0) << "\n"
       << "s0 * s1 = " << s0 * s1 << "\n"
       << "s0 * s1 * s2 = " << s0 * s1 * s2 << "\n";
  
  perm p(s0 * s1 * s2 * s3);

  cout << "s0 * s1 * s2 * s3 = " << p << "\n"
       << "last^-1 = " << (p^-1) << "\n"
       << "last^3 = " << (p^3) << "\n";
}

// }}}

#endif // ifdef TEST
C_END(perm);

// {{{ Local Variables

// Local Variables:
// mode: c++
// folded-file: t
// End:

// }}}
