WARNING;
// Copyright (C) 2000 Paul Martin and Dave Woodcock. All rights reserved.
// $Id: hash.m4,v 1.2 2000/11/13 06:16:57 dave Exp $
H_BEGIN(hash);

unsigned int hash(int, int, int);

H_END(hash);
C_BEGIN(hash);

#include <iostream.h>

// {{{ Description

// This is a hash function which takes an integer and produces an
// r-bit integer. The idea is to view the set of r-bit integers as the
// finite field GF(2^r) = GF(2)[x] / <CP(2,r)>, where CP(2,r) is the
// r^th Conway polynomial over GF(2). This is a primitive polynomial,
// i.e. the image in GF(2^r) of x is a generator of the multiplicative
// group. Thus multiplication by x is a permutation of the nonzero
// elements of GF(2^r).

// Write CP(2,r) = a_0 + a_1x + ... + a_(r-1)x^(r-1) + x^r.
// The matrix of multiplication by x with respect to the basis
// x^(r-1), x^(r-2), ... , 1 is

//     a_(r-1) 1 0 ... 0
//     a_(r-2) 0 1 ... 0

//     a_1     0 0 ... 1
//     a_0     0 0 ... 0

// If we identify GF(2^r) with r-bit integers via
//     x^i = 2^(r-1-i)
// multiplication by x becomes the function on r-bit integers v
// given by

//     if (v & 1)
//        return (v >> 1) ^ mask;
//     else
//        return v >> 1;

// where

//     mask = a_(r-1) + 2a_(r-2) + ... + 2^(r-1)a_0

// Iterating this function gives a pseudo-random number generator on r
// bits.

// We choose the reverse order so that we rotate downwards, with the
// unwanted bit falling off the bottom. If we used the usual order, we
// would rotate upwards, so would have to perform an extra bitwise and
// to kill off the unwanted top bit.

// We modify this by replacing the test by
//     if ((v & 1) ^ bit)
// where bit is a bit of the number to be hashed, and
// interating over all such bits.

// }}}
// {{{ reverseConwayMask

// See above remarks for the significance of reverseConwayMask.

// These numbers were produced by the following Gap program:
//      for i in [1..32] do
//        p := ConwayPol(2, i);
//        Unbind(p[i+1]);
//        p := Reversed(p);
//        Print(Sum([1..Length(p)], j -> 2^(j-1) * p[j]), ",\n");
//      od;

const unsigned int reverseConwayMask[] = {
  0, // The zeroth entry is meaningless
  1,
  3,
  6,
  12,
  20,
  54,
  96,
  184,
  272,
  984,
  1280,
  3440,
  6912,
  9536,
  22016,
  46080,
  73728,
  196768,
  466944,
  849408,
  1359872,
  2211328,
  4325376,
  9791360,
  21299200,
  53381120,
  95109120,
  175128576,
  335544320,
  1027944448,
  1207959552,
  2571173888
};

// }}}

// {{{ hash

unsigned int hash(int input,
		  int numInputBits,
		  int numOutputBits) {

  // It probably makes little sense to have numInputBits a lot greater
  // than numOutputBits.
  // Decreasing numInputBits speeds things up.

  unsigned int reg = reverseConwayMask[numOutputBits];
  unsigned int mask = reg;
  
  for (int i = 0; i < numInputBits; ++i) {
    if ((reg ^ input) & 1)
      reg = (reg >> 1) ^ mask;
    else
      reg >>= 1;
    input >>= 1;
  }
  return reg;
}

// }}}

C_END(hash);

// {{{ Local Variables

// Local Variables:
// mode: c++
// folded-file: t
// End:

// }}}
